#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
import dataclasses
import itertools
from typing import NoReturn, Optional, Union, List, Mapping, Callable

import numpy as np

from ase import Atoms
from ase.io import read, write

from ..data.array import AtomsNDArray
from .selector import AbstractSelector
from .cur import stat_str2val, boltz_selection, hist_selection


def get_metric_func(metric_name: str):
    """"""
    if metric_name == "fabs":
        metric_func = np.fabs
    elif metric_name == "max":
        metric_func = np.max
    elif metric_name == "min":
        metric_func = np.min
    else:
        raise NotImplementedError(f"Unknown metric function {metric_name}.")

    return metric_func


@dataclasses.dataclass
class PropertyItem:

    #: Property name that can be found in atoms.info or atoms.arrays.
    name: str

    #: metric config...
    metric: Union[str, List[str]] = None

    #: List of functions, min, max, average and ...
    _metric: List[Callable] = dataclasses.field(init=False, default_factory=list)

    #: Apply group selection.
    group: Optional[str] = None

    #: Sparsifiction method. [filter, sort, hist, boltz]
    sparsify: str = "filter"

    #: Whether reverse the sparsifiction behaviour.
    reverse: bool = False

    #: Property range to filter, which should be two strings or two numbers or mixed.
    range: List[float] = dataclasses.field(default_factory=lambda: [None, None])

    #: Property minimum.
    pmin: float = dataclasses.field(init=False, default=-np.inf)

    #: Property maximum.
    pmax: float = dataclasses.field(init=False, default=np.inf)

    #: Number of bins for histogram-based sparsification.
    nbins: int = 20

    #: Boltzmann temperature (eV).
    kBT: Optional[float] = None

    # expression
    # weight

    # worker_config: dataclasses.InitVar[str] = None

    def __post_init__(self):
        """"""
        # - bound
        bounds_ = self.range
        if bounds_[0] is None:
            bounds_[0] = "min"
        if bounds_[1] is None:
            bounds_[1] = "max"

        # NOTE: assert range when property values are available
        # assert bounds_[0] < bounds_[1], f"{self.name} has invalid bounds..."
        self.pmin, self.pmax = bounds_

        # - metric
        if self.metric is not None:
            if isinstance(self.metric, str):
                metric_config = [self.metric]
            else:
                # a list of metric function names
                metric_config = self.metric

            for metric_name in metric_config:
                if metric_name == "fabs":
                    metric_func = np.fabs
                elif metric_name == "max":
                    metric_func = np.max
                elif metric_name == "min":
                    metric_func = np.min
                else:
                    raise NotImplementedError(f"Unknown metric function {metric_name}.")
                self._metric.append(metric_func)
        else:
            self._metric = []

        # - sparsify
        assert self.sparsify in [
            "filter",
            "sort",
            "hist",
            "boltz",
        ], f"Unknown sparsification {self.sparsify}."

        return

    def _convert_raw_(self, raws_, weights_=None):
        """Convert raw values by the metric."""
        if len(self._metric) > 0:
            converts_ = []
            for raw_ in raws_:
                convert_ = raw_  # NOTE: copy?
                for metric_func in self._metric:
                    convert_ = metric_func(convert_)
                converts_.append(convert_)
        else:
            converts_ = raws_

        return converts_

    # def __repr__(self) -> str:
    #    """"""
    #    content = f"{self.name}:\n"
    #    content += f"  range: {self.pmin} - {self.pmax}\n"
    #    content += f"  sparsify: {self.sparsify}\n"

    #    return content


class PropertySelector(AbstractSelector):
    """Select structures based on structural properties.

    Each structure (trajectory) is represented by a float property.

    """

    name = "property"

    default_parameters = dict(
        mode="stru",
        properties=[],
        worker=None,  # compute properties on-the-fly
        number=[4, 0.2],
    )

    def __init__(self, directory="./", *args, **kwargs) -> None:
        """"""
        super().__init__(directory, *args, **kwargs)

        assert self.mode in ["stru", "traj"], f"Unknown selection mode {self.mode}."

        # - convert properties
        prop_items = []
        for name, params in self.properties.items():
            prop_item = PropertyItem(name=name, **params)
            prop_items.append(prop_item)
        self._prop_items = prop_items

        return

    def _mark_structures(self, data: AtomsNDArray, *args, **kwargs) -> None:
        """Return selected indices."""
        # - get property values
        #   NOTE: properties should be pre-computed...

        for prop_item in self._prop_items:
            self._print(str(prop_item))

            # - group markers
            if self.axis is None:
                marker_groups = dict(all=data.markers)
            else:
                marker_groups = {}
                for k, v in itertools.groupby(data.markers, key=lambda x: x[self.axis]):
                    if k in marker_groups:
                        marker_groups[k].extend(list(v))
                    else:
                        marker_groups[k] = list(v)
            self._debug(f"marker_groups: {marker_groups}")

            if prop_item.group is None:
                selected_markers = self._mark_group_separate(
                    data, prop_item, marker_groups
                )
            else:
                selected_markers = self._mark_group_represent(
                    data, prop_item, marker_groups
                )

            data.markers = np.array(selected_markers)

            if len(selected_markers) == 0:
                break

        return

    def _mark_group_represent(self, data, prop_item: PropertyItem, marker_groups):
        """Mark a group of structures based on a representative structure's property."""

        metric_func = get_metric_func(prop_item.group)

        rep_counter = 0
        rep_groups = []  # data for representative groups
        for grp_name, curr_markers in marker_groups.items():
            curr_frames = data.get_marked_structures(curr_markers)
            curr_nframes = len(curr_frames)

            if curr_nframes > 0:
                curr_values = self._extract_property(curr_frames, prop_item)
                metric_val = metric_func(curr_values)
                # FIXME: how to find index if structures with same properties?
                rep_frame = curr_frames[curr_values.index(metric_val)]
                rep_groups.append((grp_name, rep_counter, rep_frame, metric_val))
                rep_counter += 1
            else:
                ...
        rep_groups = sorted(rep_groups, key=lambda x: x[3])

        rep_frames = [x[2] for x in rep_groups]

        selected_markers = []
        scores, selected_indices = self._sparsify(prop_item, rep_frames)
        self._print(f"number of groups: {len(selected_indices)}")

        _counter = 0
        for grp_name, rep_index, _, _ in rep_groups:
            if rep_index in selected_indices:
                curr_selected_markers = marker_groups[grp_name]
                selected_markers.extend(curr_selected_markers)
                curr_score = scores[selected_indices.index(rep_index)]
                curr_selected_frames = data.get_marked_structures(curr_selected_markers)
                for a in curr_selected_frames:
                    a.info["score"] = curr_score
                num_curr_frames = len(curr_selected_frames)
                self._debug(f"group: {grp_name} -> {num_curr_frames}")
                _counter += num_curr_frames
            else:
                ...

        assert _counter == len(selected_markers)

        return selected_markers

    def _mark_group_separate(self, data, prop_item: PropertyItem, marker_groups):
        """Mark a group of structures based on a structure's own property."""
        selected_markers = []
        for grp_name, curr_markers in marker_groups.items():
            curr_frames = data.get_marked_structures(curr_markers)
            curr_nframes = len(curr_frames)

            # --
            if curr_nframes > 0:
                scores, selected_indices = self._sparsify(prop_item, curr_frames)
                self._print(f"number of structures: {len(selected_indices)}")
                curr_selected_markers = [curr_markers[i] for i in selected_indices]
                selected_markers.extend(curr_selected_markers)

                # - add score to atoms
                for score, i in zip(scores, selected_indices):
                    curr_frames[i].info["score"] = score

            else:
                ...

        return selected_markers

    def _extract_property(self, frames: List[Atoms], prop_item: PropertyItem):
        """Extract property values from frames.

        Returns:
            property values: List[float] or 1d-np.array.

        """
        prop_vals = []
        for atoms in frames:
            if prop_item.name == "atomic_energy":
                # TODO: move this part to PropertyItem?
                energy = atoms.get_potential_energy()
                natoms = len(atoms)
                atoms_property = energy / natoms
            elif prop_item.name == "energy":
                energy = atoms.get_potential_energy()
                atoms_property = energy
            elif prop_item.name == "forces":
                forces = atoms.get_forces(apply_constraint=True)
                atoms_property = forces
            else:
                # -- any property stored in atoms.info
                #    e.g. max_devi_f
                atoms_property = atoms.info.get(prop_item.name, None)
                if atoms_property is None:
                    atoms_property = atoms.arrays.get(prop_item.name, None)
                if atoms_property is None:
                    raise KeyError(f"{prop_item.name} does not exist.")
            prop_vals.append(atoms_property)
        prop_vals = prop_item._convert_raw_(prop_vals)

        return prop_vals

    def _statistics(self, prop_item: PropertyItem, prop_vals):
        """"""
        # - here are all data
        npoints = len(prop_vals)
        # pmax, pmin, pavg = np.max(prop_vals), np.min(prop_vals), np.mean(prop_vals)
        # pstd = np.sqrt(np.var(prop_vals-pavg))
        pmax = stat_str2val("max", prop_vals)
        pmin = stat_str2val("min", prop_vals)

        pavg = stat_str2val("avg", prop_vals)
        pstd = stat_str2val("svar", prop_vals)

        # NOTE: convert range to pmin and pmax
        #       hist data only in the range
        prop_item.pmin = stat_str2val(prop_item.pmin, prop_vals)
        prop_item.pmax = stat_str2val(prop_item.pmax, prop_vals)
        if prop_item.pmax < prop_item.pmin:
            prop_item.pmax = prop_item.pmin

        hist_max, hist_min = prop_item.pmax, prop_item.pmin
        # if hist_max == np.inf:
        #    hist_max = pmax
        # if hist_min == -np.inf:
        #    hist_min = pmin

        bins = np.linspace(hist_min, hist_max, prop_item.nbins, endpoint=False).tolist()
        bins.append(hist_max)
        hist, bin_edges = np.histogram(prop_vals, bins=bins, range=[hist_min, hist_max])

        # - output
        content = f"# Property {prop_item.name}\n"
        content += "# min {:<12.4f} max {:<12.4f}\n".format(pmin, pmax)
        content += "# avg {:<12.4f} std {:<12.4f}\n".format(pavg, pstd)
        content += "# histogram of {} points in the range (npoints: {})\n".format(
            np.sum(hist), npoints
        )
        content += f"# min {prop_item.pmin:<12.4f} max {prop_item.pmax:<12.4f}\n"
        for x, y in zip(hist, bin_edges[:-1]):
            content += "{:>12.4f}  {:>12d}\n".format(y, x)
        content += "{:>12.4f}  {:>12s}\n".format(bin_edges[-1], "-")

        with open(
            self.info_fpath.parent
            / (self.info_fpath.stem + f"-{prop_item.name}-stat.txt"),
            "w",
        ) as fopen:
            fopen.write(content)

        for l in content.split("\n"):
            self._print(l)

        return

    def _sparsify(self, prop_item: PropertyItem, frames: List[Atoms]):
        """"""
        # -- each structure is represented by one float value
        #    get per structure values
        prop_vals = self._extract_property(frames, prop_item)
        # -- give statistics of this property
        self._statistics(prop_item, prop_vals)

        nframes = len(frames)

        curr_indices, scores = [], []
        if prop_item.sparsify == "filter":
            # -- select current property
            # TODO: possibly use np.where to replace this code
            if not prop_item.reverse:
                for i in range(nframes):
                    if prop_item.pmin <= prop_vals[i] <= prop_item.pmax:
                        curr_indices.append(i)
            else:
                for i in range(nframes):
                    if prop_item.pmin > prop_vals[i] and prop_vals[i] > prop_item.pmax:
                        curr_indices.append(i)
            scores = [prop_vals[i] for i in curr_indices]
        elif prop_item.sparsify == "sort":
            numbers = list(range(nframes))
            sorted_numbers = sorted(numbers, key=lambda i: prop_vals[i])

            num_fixed = self._parse_selection_number(nframes)
            if not prop_item.reverse:
                curr_indices = sorted_numbers[:num_fixed]
            else:
                curr_indices = sorted_numbers[-num_fixed:]
            scores = [prop_vals[i] for i in curr_indices]
        elif prop_item.sparsify == "hist":
            num_fixed = self._parse_selection_number(nframes)
            prev_indices = list(range(nframes))
            scores, curr_indices = hist_selection(
                prop_item.nbins,
                prop_item.pmin,
                prop_item.pmax,
                [prop_vals[i] for i in prev_indices],
                prev_indices,
                num_fixed,
                self.rng,
            )
        elif prop_item.sparsify == "boltz":
            num_fixed = self._parse_selection_number(nframes)
            prev_indices = list(range(nframes))
            scores, curr_indices = boltz_selection(
                prop_item.kBT,
                [prop_vals[i] for i in prev_indices],
                prev_indices,
                num_fixed,
                self.rng,
            )
        else:
            # NOTE: check sparsifiction method in PropertyItem
            ...

        return scores, curr_indices


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings

import numpy as np 
import numpy.ma as ma

from ase import Atoms 
from ase.io import read, write 

from ase.neighborlist import NeighborList 
from ase.neighborlist import natural_cutoffs 

def use_neighbors():

    #covlent_cutoffs = natural_cutoffs(atoms) 
    #nl = NeighborList(covlent_cutoffs)
    #nl.update(atoms)
    #indices, offsets = nl.get_neighbors(0)

    pass 


def cut_local_environment(atoms, centre_index, cutoff_radius=8.0, box_length=20.): 
    """""" 
    # get necessities 
    cell = atoms.cell 
    frac_poses = atoms.get_scaled_positions() 
    numbers = list(atoms.get_atomic_numbers())

    centre_frac_pos = frac_poses[centre_index] 
    centre_cart_pos = np.dot(frac_poses[centre_index], cell)
    
    # form a 3x3 box 
    offsets = np.array(
        [
            [0, 0, 0], [1, 0, 0], [-1, 0, 0], 
            [0, 1, 0], [1, 1, 0], [-1, 1, 0], 
            [0, -1, 0], [1, -1, 0], [-1, -1, 0], 
            [0, 0, 1], [1, 0, 1], [-1, 0, 1], 
            [0, 1, 1], [1, 1, 1], [-1, 1, 1], 
            [0, -1, 1], [1, -1, 1], [-1, -1, 1], 
            [0, 0, -1], [1, 0, -1], [-1, 0, -1], 
            [0, 1, -1], [1, 1, -1], [-1, 1, -1], 
            [0, -1, -1], [1, -1, -1], [-1, -1, -1]
        ]
    )
    
    extended_frac_poses = np.concatenate(
        (offsets.reshape(27,1,3)+frac_poses.reshape(1,-1,3)), axis=0
    )
    #print(extended_frac_poses.shape)
    extended_numbers = np.array(numbers*27)
    #print(extended_numbers)
    
    # calculate interatomic distances 
    frac_vectors = extended_frac_poses - centre_frac_pos 
    distances = np.linalg.norm(np.dot(frac_vectors, cell), axis=1)
    
    masked_distances = ma.masked_greater_equal(distances, cutoff_radius)
    #print(masked_distances.mask)
    
    # form a cluster 
    dis_mask = masked_distances.mask
    valid_positions = np.dot(
        np.array([extended_frac_poses[i] for i, b in enumerate(dis_mask) if not b]), cell
    )
    valid_positions = valid_positions - centre_cart_pos + np.ones(3)*box_length/2. # centre the cluster 
    #print(valid_positions.shape)
    valid_numbers = ma.array(extended_numbers, mask=masked_distances.mask).compressed() 
    #print(valid_numbers.shape)
    if box_length < 2*(cutoff_radius+5.):
        warnings.warn('The box may be too small for the cluster.', UserWarning)
    cluster_atoms = Atoms(
        numbers=valid_numbers, positions=valid_positions, cell=np.eye(3)*box_length
    )
    

    return cluster_atoms

if __name__ == '__main__': 
    # this cuts the local environment of the first atom 
    atoms = read('/users/40247882/repository/gdpx/templates/structures/surf-100_opt.xyz') 
    centre_index = 0
    cluster_atoms = cut_local_environment(atoms, centre_index, cutoff_radius=7.0, box_length=20.) 
    #write('cut.xsd', cluster_atoms)
    write('cut.xyz', cluster_atoms)



#!/usr/bin/env python3
# -*- coding: utf-8 -*


import dataclasses
import pathlib
from typing import Union

import numpy as np
from scipy.interpolate import griddata, interp1d
import matplotlib.pyplot as plt
try:
    plt.style.use("presentation")
except:
    ...

from .string import AbstractStringReactor, StringReactorSetting


def interpolate_string(a, b, t, npts: int=5, concat: bool=False):
    """Inteerpolate a 2D string."""
    # - use transition state
    if concat:
        npts_1 = int(npts/2.)
        npts_2 = npts - npts_1
        pts = np.vstack(
            [
                np.hstack([np.linspace(a[0], t[0], npts_1), np.linspace(t[0], b[0], npts_2)]),
                np.hstack([np.linspace(a[1], t[1], npts_1), np.linspace(t[1], b[1], npts_2)])
            ]
        ).T
    else:
        pts = np.vstack(
            [np.linspace(a[0], b[0], npts), np.linspace(a[1], b[1], npts)]
        ).T

    return pts

def optimise_string(pts, gridpoints, z, gradx, grady, steps=100, stepsize=0.01):
    """"""
    npts = pts.shape[0]
    for i in range(steps):
        # - find gradient interpolation
        Dx = griddata(gridpoints, gradx.flatten(), (pts[:, 0], pts[:, 1]), method="linear")
        Dy = griddata(gridpoints, grady.flatten(), (pts[:, 0], pts[:, 1]), method="linear")
        h = np.amax(np.sqrt(np.square(Dx) + np.square(Dy)))
        
        # - evolve
        pts -= stepsize * np.vstack([Dx, Dy]).T / h

        # - reparameterize
        arclength = np.hstack([0, np.cumsum(np.linalg.norm(pts[1:] - pts[:-1], axis=1))])
        arclength /= arclength[-1]
        pts = np.vstack(
            [
                interp1d(arclength, pts[:, 0])(np.linspace(0, 1, npts)), 
                interp1d(arclength, pts[:, 1])(np.linspace(0, 1, npts))
            ]
        ).T

        # - save history?
        if i % 10 == 0:
            print(i, np.sum(griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear")))

    return pts


@dataclasses.dataclass
class ZeroStringReactorSetting(StringReactorSetting):

    backend: str = "grid"

    def __post_init__(self):
        """"""

        return
    
    def get_run_params(self, *args, **kwargs):
        """"""

        return


class ZeroStringReactor(AbstractStringReactor):

    name: str = "zero"

    def __init__(self, calc=None, params={}, ignore_convergence=False, directory="./", *args, **kwargs) -> None:
        """"""
        self.calc = calc
        if self.calc is not None:
            self.calc.reset()

        self.ignore_convergence = ignore_convergence

        self.directory = directory
        self.cache_nebtraj = self.directory/self.traj_name

        # - parse params
        self.setting = ZeroStringReactorSetting(**params)
        self._debug(self.setting)

        return
    
    def run(self, structures, read_cache: bool=True, *args, **kwargs):
        """"""
        self.directory.mkdir(parents=True, exist_ok=True)

        # - get points
        start_points = [a.get_positions() for a in structures]
        start_points = np.array(start_points).reshape(-1, 3)

        a, b = start_points
        nimages = self.setting.nimages
        images = np.vstack(
            [np.linspace(a[0], b[0], nimages), np.linspace(a[1], b[1], nimages)]
        ).T

        images = optimise_string(
            images, self.calc._points, self.calc._z, 
            self.calc._gradients[0], self.calc._gradients[1],
            steps=100, stepsize=0.01
        )
        energies = griddata(
            self.calc._points, self.calc._z.flatten(), 
            (images[:, 0], images[:, 1]), method="linear"
        )[:, np.newaxis]

        np.savetxt(
            self.directory/"pmf.dat", np.hstack([np.linspace(0, 1, nimages)[:, np.newaxis], images, energies]), 
            fmt="%8.4f"
        )

        # - vis
        # --
        fig, ax = plt.subplots(1, 1, figsize=(12, 8))
        ax.plot(
            np.linspace(0, 1, nimages), energies, marker="o"
        )
        ax.set_ylabel("Free Energy [eV]")
        ax.set_xlabel("Reaction Coordinate")
        fig.savefig(self.directory/"pmf.png")

        # --
        fig, ax = plt.subplots(1, 1, figsize=(12, 8))

        ax.set_xlabel("CV 1")
        ax.set_ylabel("CV 2")
        #ax.set_xlim(0., 0.33)
        #ax.set_ylim(0., 0.50)
        x, y, z = self.calc._x, self.calc._y, self.calc._z
        cntr = ax.contourf(
            x, y, z, levels=10,
            #x, np.where(y < 0., y+1.0, y), z, levels=10,
            #cmap="RdBu"
        )
        fig.colorbar(cntr, ax=ax, label="Free Energy [eV]")
        cn = ax.contour(
            x, y, z, #cmap="RdBu"
            #x, np.where(y < 0., y, y+1.0), z, levels=10,
        )
        plt.clabel(cn, inline=True, fontsize=10)

        ax.plot(
            images[:, 0], images[:,1], linestyle="--"
        )
        fig.savefig(self.directory/"fes.png", transparent=False)

        return
    
    def read_convergence(self, *args, **kwargs) -> bool:
        """"""
        converged = super().read_convergence(*args, **kwargs)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib

import numpy as np

from ase.io import read, write


def cut_frames(p, beg_step=0, end_step=5000000, period=500):
    """"""
    frames = read(p, ":")

    #return frames[int(beg_step/period):int(end_step/period)+1]
    return frames[beg_step:end_step+1]

def get_colvars(p, beg_step, end_step, period=500):
    """"""
    colvars = np.loadtxt(p)

    #return colvars[:int(maxstep/period), 1:3] # pos.a, pos.b
    #return colvars[int(beg_step/period)+1:int(end_step/period)+1, 4:6] # sx, sy
    return colvars[beg_step:end_step+1, 4:6] # sx, sy

def get_frames(p, beg_step=0, end_step=5000000, period=500):
    """"""
    p = pathlib.Path(p)
    frames = cut_frames(
        p/"traj.dump", beg_step=0, end_step=end_step-beg_step, period=period
    )
    nframes = len(frames)
    print(f"nframes: {nframes}")

    # NOTE: COLVAR has accumulated steps
    colvars = get_colvars(p/"COLVAR", beg_step, end_step, period=period)
    print(f"colvars: {colvars.shape}")
    #print(colvars)

    for a, cv in zip(frames, colvars):
        a.info["colvar"] = "{},{}".format(*cv)
    #write(p/"traj.xyz", frames)

    return frames

#frames = []
#curr_frames = get_frames("./0000.run", end_step=20000, period=250)
#frames.extend(curr_frames)
#
#curr_frames = get_frames("./0001.run", beg_step=20001, end_step=40001, period=255)
#frames.extend(curr_frames[1:])
#
#curr_frames = get_frames("./0002.run", beg_step=40002, end_step=60002, period=255)
#frames.extend(curr_frames[1:])
#
#write("traj.xyz", frames)

frames = read("traj.xyz", ":")
colvars = np.array(
    [a.info["colvar"] for a in frames], dtype=np.float32
)

nframes = len(frames)
print(f"nframes: {nframes}")
print(f"colvars: {colvars.shape}")

coordinates = np.loadtxt("./0002.run/fes/path.txt")
print(coordinates)

nbins = 10
size = 24 # 128
eps = 0.05 # eps = 0.05

groups = [[] for _ in range(nbins)]
for i, (com_x, com_y) in enumerate(colvars):
    for j, (x, y) in enumerate(coordinates):
        if x-eps < com_x < x+eps and y-eps < com_y < y+eps:
            #print(com_x, com_y)
            groups[j].append(i)
print([len(_) for _ in groups])
print(sum([len(_) for _ in groups]))

content = ""
for i, x in enumerate(groups):
    content += " ".join([str(_) for _ in x]) + "\n"

with open("eps.txt", "w") as fopen:
    fopen.write(content)

rng = np.random.default_rng(1112)

selected_indices = []
for g in groups:
    s = rng.choice(g, size=size, replace=False).tolist()
    print(s)
    selected_indices.extend(s)
nselected = len(selected_indices)
print(f"nselected: {nselected}")

selected_frames = [frames[i] for i in selected_indices]
write("frames_to_reweight.xyz", selected_frames)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools

from typing import List, Mapping

from ase import Atoms


def find_product(atoms: Atoms, reactants: List[List[int]], grid=[1,1,0], radii_multi=1.0, skin=0.0) -> List[List[int]]:
    """Find if there were a product from input reactants."""
    valid_indices = list(itertools.chain.from_iterable(reactants))

    # - create local graph
    covalent_radii = natural_cutoffs(atoms, radii_multi)
    nl = NeighborList(
        covalent_radii, 
        skin = skin, sorted=False,
        self_interaction=False, 
        bothways=True
    )
    nl.update(atoms)

    #print([covalent_radii[i] for i in valid_indices])

    graph = nx.Graph()
    
    #grid = [1,1,0] # for surface
    # -- add nodes
    for centre_idx in valid_indices:
        for x, y, z in grid_iterator(grid):
            graph.add_node(
                node_symbol(atoms[centre_idx].symbol, centre_idx, (x,y,z)),
                index=centre_idx
            )

    # -- add edges
    for centre_idx in valid_indices:
        for x, y, z in grid_iterator(grid):
            nei_indices, nei_offsets = nl.get_neighbors(centre_idx)
            for nei_idx, offset in zip(nei_indices, nei_offsets):
                if nei_idx in valid_indices:
                    # NOTE: check if neighbour is in the grid space
                    #       this is not the case when cutoff is too large
                    ox, oy, oz = offset
                    if not (-grid[0] <= ox + x <= grid[0]):
                        continue
                    if not (-grid[1] <= oy + y <= grid[1]):
                        continue
                    if not (-grid[2] <= oz + z <= grid[2]):
                        continue
                    # ---
                    graph.add_edge(
                        node_symbol(atoms[centre_idx].symbol, centre_idx, (x,y,z)),
                        node_symbol(atoms[nei_idx].symbol, nei_idx, (x+ox,y+oy,z+oz))
                    )
                else:
                    ...
    
    #plot_graph(graph, "xxx.png")

    # - find products
    reax_nodes = [node_symbol(atoms[i].symbol, i, (0,0,0)) for i in valid_indices]
    reax_graphs = nx.subgraph(graph, reax_nodes)

    prod_graphs = [reax_graphs.subgraph(c) for c in nx.connected_components(reax_graphs)]

    products = [[unpack_node_name(u)[1] for u in g.nodes()] for g in prod_graphs]

    return products

from ase.formula import Formula
def find_molecules(atoms: Atoms, valid_indices: List[int], grid=[1,1,0], radii_multi=1.0, skin=0.0) -> Mapping[str,List[List[int]]]:
    """Find if there were a product from input reactants."""
    #valid_indices = list(chain.from_iterable(reactants))

    # - create local graph
    covalent_radii = natural_cutoffs(atoms, radii_multi)
    nl = NeighborList(
        covalent_radii, 
        skin = skin, sorted=False,
        self_interaction=False, 
        bothways=True
    )
    nl.update(atoms)

    #print([covalent_radii[i] for i in valid_indices])

    graph = nx.Graph()
    
    #grid = [1,1,0] # for surface
    # -- add nodes
    for centre_idx in valid_indices:
        for x, y, z in grid_iterator(grid):
            graph.add_node(
                node_symbol(atoms[centre_idx].symbol, centre_idx, (x,y,z)),
                index=centre_idx
            )

    # -- add edges
    for centre_idx in valid_indices:
        for x, y, z in grid_iterator(grid):
            nei_indices, nei_offsets = nl.get_neighbors(centre_idx)
            for nei_idx, offset in zip(nei_indices, nei_offsets):
                if nei_idx in valid_indices:
                    # NOTE: check if neighbour is in the grid space
                    #       this is not the case when cutoff is too large
                    ox, oy, oz = offset
                    if not (-grid[0] <= ox + x <= grid[0]):
                        continue
                    if not (-grid[1] <= oy + y <= grid[1]):
                        continue
                    if not (-grid[2] <= oz + z <= grid[2]):
                        continue
                    # ---
                    graph.add_edge(
                        node_symbol(atoms[centre_idx].symbol, centre_idx, (x,y,z)),
                        node_symbol(atoms[nei_idx].symbol, nei_idx, (x+ox,y+oy,z+oz))
                    )
                else:
                    ...
    
    #plot_graph(graph, "xxx.png")

    # - find products
    reax_nodes = [node_symbol(atoms[i].symbol, i, (0,0,0)) for i in valid_indices]
    reax_graphs = nx.subgraph(graph, reax_nodes)

    prod_graphs = [reax_graphs.subgraph(c) for c in nx.connected_components(reax_graphs)]

    products = [[unpack_node_name(u)[1] for u in g.nodes()] for g in prod_graphs]

    # - get formula
    fragments = {}
    for atomic_indices in products:
        symbols = [atoms[i].symbol for i in atomic_indices]
        formula = Formula.from_list(symbols).format("hill")
        if formula in fragments:
            fragments[formula].append(atomic_indices)
        else:
            fragments[formula] = [atomic_indices]

    return fragments 


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import os
import copy
import dataclasses
import time
import shutil
import warnings
import pathlib
import tarfile
import tempfile
import traceback
from pathlib import Path
from typing import List, Tuple

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.calculator import FileIOCalculator, EnvironmentError
from ase.calculators.singlepoint import SinglePointCalculator
from ase.geometry import find_mic

from ..builder.constraints import parse_constraint_info
from .driver import AbstractDriver, DriverSetting


"""Driver and calculator of LaspNN.

Output files by LASP NVE-MD are 
    allfor.arc  allkeys.log  allstr.arc  firststep.restart 
    md.arc  md.restart  vel.arc

"""

def is_number(s):
    """"""
    try:
        float(s)
        return True
    except ValueError:
        ...

    return False

def compare_trajectory_continuity(t0, t1):
    """Compare positions."""
    a0, a1 = t0[-1], t1[0]
    cell = a0.get_cell(complete=True)
    shift = a0.positions - a1.positions
    curr_vectors, curr_distances = find_mic(shift, cell, pbc=True)

    # Due to the floating point precision, arc -> atoms may lead to
    # position inconsistent after 8 decimals...
    return np.allclose(
        curr_vectors, np.zeros(curr_vectors.shape), 
        #rtol=1e-05, atol=1e-08, equal_nan=False
        rtol=1e-04, atol=1e-06, equal_nan=False
    )


class LaspEnergyError(Exception):
    """Error due to the failure of LASP energy.

    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)

def read_laspset(train_structures):
    """Read LASP TrainStr.txt and TrainFor.txt files."""
    train_structures = Path(train_structures)
    frames = []

    all_energies, all_forces, all_stresses = [], [], []

    # - TrainStr.txt
    # TODO: use yield
    with open(train_structures, "r") as fopen:
        while True:
            line = fopen.readline()
            if line.strip().startswith("Start one structure"):
                # - energy
                line = fopen.readline()
                energy = float(line.strip().split()[-2])
                all_energies.append(energy)
                # - natoms
                line = fopen.readline()
                natoms = int(line.strip().split()[-1])
                # skip 5 lines, symbol info and training weights
                #skipped_lines = [fopen.readline() for i in range(5)]
                # - cell
                cell = []
                for _ in range(1000):
                    lat_line = fopen.readline()
                    if lat_line.strip().startswith("lat"):
                        cell.append(lat_line.strip().split()[1:])
                    if len(cell) == 3:
                        break
                else:
                    raise RuntimeError("Failed to read lattice.")
                cell = np.array(cell, dtype=float)
                # - symbols, positions, and charges
                anumbers, positions, charges = [], [], []
                for i in range(natoms):
                    data = fopen.readline().strip().split()[1:]
                    anumbers.append(int(data[0]))
                    positions.append([float(x) for x in data[1:4]])
                    charges.append(float(data[-1]))
                atoms = Atoms(numbers=anumbers, positions=positions, cell=cell, pbc=True)
                assert fopen.readline().strip().startswith("End one structure")
                frames.append(atoms)
                #break
            if not line:
                break
    
    # - TrainFor.txt
    train_forces = train_structures.parent / "TrainFor.txt"
    with open(train_forces, "r") as fopen:
        while True:
            line = fopen.readline()
            if line.strip().startswith("Start one structure"):
                # - stress, voigt order
                stress = np.array(fopen.readline().strip().split()[1:], dtype=float)
                # - symbols, forces
                anumbers, forces = [], []
                line = fopen.readline()
                while True:
                    if line.strip().startswith("force"):
                        data = line.strip().split()[1:]
                        anumbers.append(int(data[0]))
                        forces.append([float(x) for x in data[1:4]])
                    else:
                        all_forces.append(forces)
                        assert line.strip().startswith("End one structure")
                        break
                    line = fopen.readline()
                #break
            if not line:
                break
    
    for i, atoms in enumerate(frames):
        calc = SinglePointCalculator(
            atoms, energy=all_energies[i], forces=all_forces[i]
        )
        atoms.calc = calc
    write(train_structures.parent / "dataset.xyz", frames)

    return frames

def read_lasp_structures(
        mdir: pathlib.Path, wdir: pathlib.Path, archive_path: pathlib.Path=None, 
        *args, **kwargs
    ) -> List[Atoms]:
    """Read simulation trajectory."""
    wdir = pathlib.Path(wdir)

    # - check if output file exists...
    if (not (wdir/"allstr.arc").exists()) and archive_path is None:
        return []

    # - get IO
    if archive_path is None:
        with open(wdir/"allstr.arc", "r") as fopen:
            stru_io = io.StringIO(fopen.read())
        afrc_io = open(wdir/"allfor.arc", "r") # atomic forces in arc format
        lout_io = open(wdir/"lasp.out", "r")
    else:
        rpath = wdir.relative_to(mdir.parent)
        stru_tarname = str(rpath/"allstr.arc")
        afrc_tarname = str(rpath/"allfor.arc")
        lout_tarname = str(rpath/"lasp.out")
        with tarfile.open(archive_path, "r:gz") as tar:
            for tarinfo in tar:
                if tarinfo.name.startswith(wdir.name):
                    if tarinfo.name == stru_tarname:
                        stru_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    elif tarinfo.name == afrc_tarname:
                        afrc_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    elif tarinfo.name == lout_tarname:
                        lout_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    else:
                        ...
                else:
                    ...
            else: # TODO: if not find target traj?
                ...

    # - parse data
    # NOTE: ASE does not support read dmol-arc from stringIO
    with tempfile.NamedTemporaryFile(mode="w", suffix=".arc") as tmp:
        tmp.write(stru_io.getvalue())
        tmp.seek(0)
        traj_frames = read(tmp.name, ":", format="dmol-arc")
    natoms = len(traj_frames[-1])

    traj_steps = []
    traj_energies = []
    traj_forces = []

    # NOTE: for lbfgs opt, steps+2 frames would be output?

    # have to read last structure
    while True:
        line = afrc_io.readline()
        if line.strip().startswith("For"):
            step = int(line.split()[1])
            traj_steps.append(step)
            # NOTE: need check if energy is a number
            #       some ill structure may result in large energy of ******
            energy_data = line.split()[3]
            try:
                energy = float(energy_data)
            except ValueError:
                energy = np.inf
                msg = "Energy is too large at {}. The structure maybe ill-constructed.".format(wdir)
                warnings.warn(msg, UserWarning)
            traj_energies.append(energy)
            # stress
            line = afrc_io.readline()
            stress = np.array(line.split()) # TODO: what is the format of stress
            # forces
            forces = []
            for j in range(natoms):
                line = afrc_io.readline()
                force_data = line.strip().split()
                if len(force_data) == 3: # expect three numbers
                    force_data_ = []
                    for x in force_data:
                        if not is_number(x):
                            force_data_.append(np.inf)
                        else:
                            force_data_.append(float(x))
                    force_data = force_data_
                else: # too large forces make out become ******
                    force_data = [np.inf]*3
                forces.append(force_data)
            forces = np.array(forces, dtype=float)
            traj_forces.append(forces)
        if line.strip() == "":
            pass
        if not line: # if line == "":
            break
    assert len(traj_frames) == len(traj_steps), "Output number is inconsistent."
    
    # - create traj
    for i, atoms in enumerate(traj_frames):
        calc = SinglePointCalculator(
            atoms, energy=traj_energies[i], forces=traj_forces[i]
        )
        atoms.calc = calc

    # check if the structure is too bad... LASP v3.3.4
    TOO_SHORT_BOND_TAG = "Warning: Minimum Structure with too short bond"
    is_badstru = False
    lines = lout_io.readlines()
    for line in lines:
        if TOO_SHORT_BOND_TAG in line:
            is_badstru = True
            break
        else:
            ...
    traj_frames[-1].info["is_badstru"] = is_badstru

    # - close IO
    stru_io.close()
    afrc_io.close()
    lout_io.close()

    return traj_frames

@dataclasses.dataclass
class LaspDriverSetting(DriverSetting):

    def __post_init__(self):
        """"""
        if self.task == "min":
            self._internals.update(
                **{
                    "explore_type": "ssw",
                    "SSW.SSWsteps": 1, # BFGS
                    "SSW.ftol": self.fmax
                }
            )

            assert self.dump_period == 1, "LaspDriver must have dump_period ==1."
        
        if self.task == "md":
            if self.tend is None:
                self.tend = self.temp
            self._internals.update(
                **{
                    "explore_type": self.md_style,
                    "Ranseed": self.velocity_seed,
                    "MD.dt": self.timestep,
                    "MD.initial_T": self.temp,
                    "MD.target_T": self.tend,
                    "nhmass": self.Tdamp,
                    "MD.target_P": self.press,
                    "MD.prmass": self.Pdamp,
                }
            )
            ...

        # - shared params

        # - special params

        return
    
    def get_run_params(self, *args, **kwargs):
        """"""
        steps_ = kwargs.get("steps", self.steps)
        fmax_ = kwargs.get("fmax", self.fmax)

        run_params = {
            "constraint": kwargs.get("constraint", self.constraint),
            "SSW.MaxOptstep": steps_
        }

        if self.task == "min":
            run_params.update(
                **{"SSW.ftol": fmax_}
            )
        
        if self.task == "md":
            timestep = self._internals["MD.dt"]
            run_params.update(
                **{
                    "MD.ttotal": timestep*steps_,
                    "MD.print_freq": self.dump_period*timestep, # freq has unit fs
                    "MD.print_strfreq": self.dump_period*timestep
                }
            )

        # - add extra parameters
        run_params.update(
            **kwargs
        )

        return run_params

class LaspDriver(AbstractDriver):

    """Driver for LASP."""

    name = "lasp"

    #: Whether accepct the bad structure due to crashed FF or SCF-unconverged DFT.
    accept_bad_structure: bool = True

    # - defaults
    default_task = "min"
    supported_tasks = ["min", "md"]

    def __init__(self, calc, params: dict, directory="./", *args, **kwargs):
        """"""
        super().__init__(calc, params, directory=directory, *args, **kwargs)

        self.setting = LaspDriverSetting(**params)

        return
    
    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """"""
        verified = super()._verify_checkpoint(*args, **kwargs)
        if verified:
            laspstr = self.directory / "allstr.arc"
            if (laspstr.exists() and laspstr.stat().st_size != 0):
                verified = True
            else:
                verified = False
        else:
            ...

        return verified

    def _irun(self, atoms: Atoms, ckpt_wdir=None, cache_traj: List[Atoms]=None, *args, **kwargs) -> None:
        """"""
        try:
            if ckpt_wdir is None: # start from the scratch
                # - init params
                run_params = self.setting.get_init_params()
                run_params.update(**self.setting.get_run_params(**kwargs))

                self.calc.set(**run_params)
                atoms.calc = self.calc

                _ = atoms.get_forces()
            else:
                # TODO: velocities?
                if cache_traj is None:
                    traj = self.read_trajectory()
                else:
                    self._debug("use cache trajectory to restart...")
                    traj = cache_traj
                nframes = len(traj)
                assert nframes > 0, "LaspDriver restarts with a zero-frame trajectory."
                atoms = traj[-1]
                target_steps = self.setting.steps
                dump_period = self.setting.dump_period
                if target_steps > 0:
                    steps = target_steps + dump_period - nframes*dump_period
                assert steps > 0, "Steps should be greater than 0."
                run_params = self.setting.get_init_params()
                run_params.update(**self.setting.get_run_params(steps=steps))

                self.calc.set(**run_params)
                atoms.calc = self.calc

                _ = atoms.get_forces()

        except Exception as e:
            self._debug(e)
            self._debug(traceback.print_exc())

        return
    
    def read_force_convergence(self, *args, **kwargs) -> bool:
        """"""
        return self.calc._is_converged()
    
    def _read_a_single_trajectory(self, wdir: pathlib.Path, archive_path: pathlib.Path, *args, **kwargs) -> List[Atoms]:
        """"""
        curr_frames = read_lasp_structures(self.directory, wdir, archive_path=archive_path)

        return curr_frames
    
    def read_trajectory(self, archive_path: pathlib.Path=None, *args, **kwargs) -> List[Atoms]:
        """Read trajectory in the current working directory."""
        prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        self._debug(f"prev_wdirs: {prev_wdirs}")

        traj_list = []
        for w in prev_wdirs:
            curr_frames = self._read_a_single_trajectory(w, archive_path)
            traj_list.append(curr_frames)
        
        # Even though arc file may be empty, the read can give a empty list...
        laspstr = self.directory / "allstr.arc"
        traj_list.append(self._read_a_single_trajectory(self.directory, archive_path))

        # -- concatenate
        traj_frames, ntrajs = [], len(traj_list)
        if ntrajs > 0:
            traj_frames.extend(traj_list[0])
            for i in range(1, ntrajs):
                if traj_list[i]: # check if the traj is a empty list
                    #assert np.allclose(traj_list[i-1][-1].positions, traj_list[i][0].positions), f"Traj {i-1} and traj {i} are not consecutive."
                    assert compare_trajectory_continuity(traj_list[i-1], traj_list[i]), f"Traj {i-1} and traj {i} are not consecutive."
                    traj_frames.extend(traj_list[i][1:])
                else:
                    ...
        else:
            ...
        
        nframes = len(traj_frames)
        self._debug(f"LASP read_trajectory nframes: {nframes}")

        # NOTE: LASP only save step info in MD simulations...
        for i in range(nframes):
            traj_frames[i].info["step"] = i*self.setting.dump_period

        return traj_frames


class LaspNN(FileIOCalculator):

    #: Calculator name.
    name: str = "LaspNN"

    #: Implemented properties.
    implemented_properties: List[str] = ["energy", "forces"]
    # implemented_propertoes = ["energy", "forces", "stress"]

    #: LASP command.
    command = "lasp"

    #: Default calculator parameters, NOTE which have ase units.
    default_parameters = {
        # built-in parameters
        "potential": "NN",
        # - general settings
        "explore_type": "ssw", # ssw, nve nvt npt rigidssw train
        "Ranseed": None,
        # - ssw
        "SSW.internal_LJ": True,
        "SSW.ftol": 0.05, # fmax
        "SSW.SSWsteps": 0, # 0 sp 1 opt >1 ssw search
        "SSW.Bfgs_maxstepsize": 0.2,
        "SSW.MaxOptstep": 0, # lasp default 300
        "SSW.output": "T",
        "SSW.printevery": "T",
        # md
        "MD.dt": 1.0, # fs
        "MD.ttotal": 0, 
        "MD.initial_T": None,
        "MD.equit": 0,
        "MD.target_T": 300, # K
        "MD.nhmass": 1000, # eV*fs**2
        "MD.target_P": 300, # 1 bar = 1e-4 GPa
        "MD.prmass": 1000, # eV*fs**2
        "MD.realmass": ".true.",
        "MD.print_freq": 10,
        "MD.print_strfreq": 10,
        # calculator-related
        "constraint": None, # str, lammps-like notation
    }

    def __init__(self, *args, label="LASP", **kwargs):
        """Init calculator.

        The potential path would be resolved.

        """
        FileIOCalculator.__init__(self, *args, label=label, **kwargs)

        # NOTE: need resolved pot path
        pot_ = {}
        pot = self.parameters.get("pot", None)
        for k, v in pot.items():
            pot_[k] = Path(v).resolve()
        self.set(pot=pot_)

        return
    
    def calculate(self, *args, **kwargs):
        """Perform the calculation."""
        FileIOCalculator.calculate(self, *args, **kwargs)

        return
    
    def write_input(self, atoms, properties=None, system_changes=None):
        """Write LASP inputs."""
        # create calc dir
        FileIOCalculator.write_input(self, atoms, properties, system_changes)

        # structure
        write(
            os.path.join(self.directory, "lasp.str"), atoms, format="dmol-arc",
            parallel=False
        )

        # check symbols and corresponding potential file
        atomic_types = set(self.atoms.get_chemical_symbols()) # TODO: sort by 

        # - potential choice
        # NOTE: only for LaspNN now
        content  = "potential {}\n".format(self.parameters["potential"])
        assert self.parameters["potential"] == "NN", "Lasp calculator only support NN now."

        content += "%block netinfo\n"
        for atype in atomic_types:
            # write path
            pot_path = Path(self.parameters["pot"][atype]).resolve()
            content += "  {:<4s} {:s}\n".format(atype, pot_path.name)
            # creat potential link
            pot_link = Path(os.path.join(self.directory,pot_path.name))
            if not pot_link.is_symlink(): # false if not exists
                pot_link.symlink_to(pot_path)
        content += "%endblock netinfo\n"

        # - atom constraint
        constraint = self.parameters["constraint"]
        mobile_text, frozen_text = parse_constraint_info(atoms, constraint)

        if frozen_text is not None:
            content += "%block fixatom\n"
            frozen_block = frozen_text.strip().split()
            for block in frozen_block:
                info = block.split(":")
                if len(info) == 2:
                    s, e = info
                else:
                    s, e = info[0], info[0]
                content += "  {} {} xyz\n".format(s, e)
            content += "%endblock fixatom\n"
        
        # - general settings
        seed = self.parameters["Ranseed"]
        if seed is None:
            seed = np.random.randint(1,10000)
        content += "{}  {}".format("Ranseed", seed)

        # - simulation task
        explore_type = self.parameters["explore_type"]
        content += "\nexplore_type {}\n".format(explore_type)
        if explore_type == "ssw":
            for key, value in self.parameters.items():
                if key.startswith("SSW."):
                    content += "{}  {}\n".format(key, value)
        elif explore_type in ["nve", "nvt", "npt"]:
            required_keys = [
                "MD.dt", "MD.ttotal", "MD.realmass", 
                "MD.print_freq", "MD.print_strfreq"
            ]
            if explore_type == "nvt":
                required_keys.extend(["MD.initial_T", "MD.target_T", "MD.equit"])
            if explore_type == "npt":
                required_keys.extend(["MD.target_P"])
            
            self.parameters["MD.ttotal"] = (
                self.parameters["MD.dt"] * self.parameters["SSW.MaxOptstep"]
            )

            for k, v in self.parameters.items():
                if k == "MD.target_P":
                    v *= 1e4 # from bar to GPa
                if k in required_keys:
                    content += "{}  {}\n".format(k, v)
        else:
            # TODO: should check explore_type in init
            pass

        with open(os.path.join(self.directory, "lasp.in"), "w") as fopen:
            fopen.write(content)

        return
    
    def read_results(self):
        """Read LASP results."""
        # have to read last structure
        traj_frames = read_lasp_structures(self.directory, self.directory)

        energy = traj_frames[-1].get_potential_energy()
        forces = traj_frames[-1].get_forces().copy()

        self.results["energy"] = energy
        self.results["forces"] = forces

        return
    
    def _is_converged(self) -> bool:
        """Check whether LASP simulation is converged."""
        converged = False
        lasp_out = Path(os.path.join(self.directory, "lasp.out" ))
        if lasp_out.exists():
            with open(lasp_out, "r") as fopen:
                lines = fopen.readlines()
            if (# NOTE: its a typo in LASP!!
                lines[-1].strip().startswith("elapse_time") or # v3.3.4
                lines[-1].strip().startswith("Elapse_time")    # v3.4.5
            ):
                converged = True

        return converged


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

from pathlib import Path

from . import AbstractPotentialManager


class LaspManager(AbstractPotentialManager):

    name = "lasp"
    implemented_backends = ["lasp"]
    valid_combinations = (
        ("lasp", "lasp"), # calculator, dynamics
        ("lasp", "ase")
    )

    def __init__(self):

        return

    def register_calculator(self, calc_params):
        """ params
            command
            directory
            pot
        """
        super().register_calculator(calc_params)

        self.calc_params["pot_name"] = self.name

        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", Path.cwd())
        atypes = calc_params.pop("type_list", [])

        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))
        
        # update to resolved paths
        self.calc_params["model"] = models
        
        pot = {}
        if len(models) == len(atypes):
            for t, m in zip(atypes,models):
                pot[t] = m
        else:
            # use first model for all types
            for t in atypes:
                pot[t] = models[0]

        self.calc = None
        if self.calc_backend == "lasp":
            from gdpx.computation.lasp import LaspNN
            self.calc = LaspNN(
                command=command, directory=directory, pot=pot,
                **calc_params
            )
        elif self.calc_backend == "lammps":
            # TODO: add lammps calculator
            pass
        else:
            raise NotImplementedError(f"{self.name} does not have {self.calc_backend}.")

        return


if __name__ == "__main__":
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .distance import DistanceHarmonicCalculator
from .plane import PlaneHarmonicCalculator


__all__ = ["DistanceHarmonicCalculator", "PlaneHarmonicCalculator"]


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import itertools
import pathlib
import re
import warnings

from typing import Any, NoReturn, Optional, List, Mapping

import numpy as np

from ase import Atoms
from ase.io import read, write

from .. import config
from ..core.register import registers
from ..core.variable import Variable
from ..core.operation import Operation

from .array import AtomsNDArray
from .dataset import XyzDataloader
from .system import DataSystem


# --- variable

@registers.variable.register
class TempdataVariable(Variable):

    def __init__(self, system_dirs, *args, **kwargs):
        """"""
        initial_value = self._process_dataset(system_dirs)
        super().__init__(initial_value)

        return
    
    def _process_dataset(self, system_dirs: List[str]):
        """"""
        #system_dirs.sort()
        system_dirs = [pathlib.Path(s) for s in system_dirs]

        dataset = []
        for s in system_dirs:
            prefix, curr_frames = s.name, []
            xyzpaths = sorted(list(s.glob("*.xyz")))
            for p in xyzpaths:
                curr_frames.extend(read(p, ":"))
            dataset.append([prefix, curr_frames])

        return dataset


@registers.variable.register
class NamedTempdataVariable(Variable):

    def __init__(self, system_dirs, *args, **kwargs):
        """"""
        initial_value = self._process_dataset(system_dirs)
        super().__init__(initial_value)

        return
    
    def _process_dataset(self, system_dirs: List[str]):
        """"""
        #system_dirs.sort()
        system_dirs = [pathlib.Path(s) for s in system_dirs]

        data_systems = []
        for s in system_dirs:
            config._debug(str(s))
            d = DataSystem(directory=s)
            data_systems.append(d)

        return data_systems


@registers.variable.register
class DatasetVariable(Variable):

    def __init__(self, name, directory="./", *args, **kwargs):
        """"""
        dataset = registers.create("dataloader", name, convert_name=True, **kwargs)
        super().__init__(initial_value=dataset, directory=directory)

        return


# --- operation
@registers.operation.register
class assemble(Operation):

    def __init__(self, variable, directory="./", **kwargs) -> None:
        """"""
        vkwargs = {} # variable non-variable/operation kwargs
        node_names, input_nodes = [], []
        for k, v in kwargs.items():
            if isinstance(v, (Variable, Operation)):
                node_names.append(k)
                input_nodes.append(v)
            else:
                vkwargs[k] = v
        super().__init__(input_nodes, directory)

        self.variable = variable
        self.vkwargs = vkwargs
        self.node_names = node_names

        return
    
    def forward(self, *outputs):
        """"""
        super().forward()

        # NOTE: check whether dependant nodes all have valid outputs
        is_finished = True
        for o in outputs:
            if o is None:
                is_finished = False
                break
        
        if is_finished:
            params = copy.deepcopy(self.vkwargs)
            params.update({k: v for k, v in zip(self.node_names, outputs)})
            variable = registers.create("variable", self.variable, **params)
            ret = variable.value
        else:
            ret = None

        self.status = "finished" if is_finished else "unfinished"

        return ret

@registers.operation.register
class seqrun(Operation):

    def __init__(self, nodes, *args, **kwargs) -> NoReturn:
        """"""
        super().__init__(nodes)

        return
    
    def forward(self, *outputs):
        """"""
        super().forward()

        return

@registers.operation.register
class chain(Operation):

    """Merge arbitrary nodes' outputs into one list.
    """

    status = "finished" # Always finished since it is not time-consuming

    def __init__(self, nodes, *args, **kwargs) -> NoReturn:
        """"""
        super().__init__(nodes)

        # - some operation parameters

        return
    
    def forward(self, *outputs):
        """"""
        super().forward()
        self._debug(f"chain outputs: {outputs}")

        return list(itertools.chain(*outputs))

@registers.operation.register
class map(Operation):

    """Give each input node a name and construct a dict.

    This is useful when validating structures from different systems.

    """

    def __init__(self, nodes, names, directory="./") -> NoReturn:
        """"""
        super().__init__(nodes, directory)

        assert len(nodes) == len(names), "Numbers of nodes and names are inconsistent."
        self.names = names

        return
    
    def forward(self, *outputs):
        """"""
        super().forward()

        ret = {}
        for k, v in zip(self.names, outputs):
            ret[k] = v
        
        self._debug(f"map ret: {ret}")

        # NOTE: some dependant nodes may not finish
        self.status = "finished"

        return ret

@registers.operation.register
class zip_nodes(Operation):

    status = "finished"

    def __init__(self, nodes, directory="./") -> None:
        """"""
        super().__init__(input_nodes=nodes, directory=directory)

        return

    def forward(self, *outputs):
        """"""
        super().forward()

        ret = [list(x) for x in zip(*outputs)]

        return ret

@registers.operation.register
class list_nodes(Operation):

    status = "finished"

    def __init__(self, nodes, directory="./") -> None:
        """"""
        super().__init__(input_nodes=nodes, directory=directory)

        return

    def forward(self, *outputs):
        """"""
        super().forward()

        ret = list(outputs)

        return ret


@registers.operation.register
class transfer(Operation):

    """Transfer worker results to target destination.
    """

    def __init__(
        self, structures, dataset, version, 
        prefix: str="", system: str="mixed", clean_info: bool=False,
        directory="./"
    ) -> None:
        """"""
        input_nodes = [structures, dataset]
        super().__init__(input_nodes=input_nodes, directory=directory)

        self.version = version

        self.prefix = prefix
        self.system = system # molecule/cluster, surface, bulk

        self.clean_info = clean_info # whether clean atoms info

        return
    
    def forward(self, frames: List[Atoms], dataset):
        """"""
        super().forward()

        if isinstance(frames, AtomsNDArray):
            frames = frames.get_marked_structures()
        nframes = len(frames)
        self._print(f"{nframes = }")

        target_dir = dataset.directory.resolve()
        self._print(f"target dir: {str(target_dir)}")

        # - check chemical symbols
        system_dict = {} # {formula: [indices]}

        # NOTE: groupby only collects contiguous data
        #       we need aggregate by ourselves
        formulae = [a.get_chemical_formula() for a in frames]
        for k, v in itertools.groupby(enumerate(formulae), key=lambda x: x[1]):
            if k not in system_dict:
                system_dict[k] = [x[0] for x in v]
            else:
                system_dict[k].extend([x[0] for x in v])
        
        # - transfer data
        acc_nframes = 0
        for formula, curr_indices in system_dict.items():
            # -- TODO: check system type
            system_type = self.system # currently, use user input one
            # -- name = description+formula+system_type
            #dirname = "-".join([self.directory.parent.name, formula, system_type])
            dirname = "-".join([self.prefix, formula, system_type])

            target_subdir = target_dir/dirname
            target_subdir.mkdir(parents=True, exist_ok=True)

            # -- save frames
            curr_frames = [frames[i] for i in curr_indices]
            curr_nframes = len(curr_frames)

            if self.clean_info:
                self._clean_frames(curr_frames)

            strname = self.version + ".xyz"
            target_destination = target_dir/dirname/strname
            if not target_destination.exists():
                write(target_destination, curr_frames)
                self._print(f"nframes {curr_nframes} -> {str(target_destination.relative_to(target_dir))}")
            else:
                #warnings.warn(f"{target_destination} exists.", UserWarning)
                self._print(f"WARN: {str(target_destination.relative_to(target_dir))} exists.")

            acc_nframes += curr_nframes

        assert nframes == acc_nframes
        
        self.status = "finished"

        return dataset
    
    def _clean_frames(self, frames: List[Atoms]):
        """"""
        for atoms in frames:
            info_keys = copy.deepcopy(list(atoms.info.keys()))
            for k in info_keys:
                if k not in ["energy", "free_energy"]:
                    del atoms.info[k]

        return


@registers.operation.register
class scope(Operation):

    def __init__(
            self, dataset, describer, 
            groups: Optional[dict]=None, subgroups: Optional[dict]=None, level: int=0,
            add_legend: bool=True, write_chemiscope=False, directory="./"
        ) -> None:
        """"""
        super().__init__(input_nodes=[dataset, describer], directory=directory)
        
        self.level = level

        if groups is None:
            groups = {"all": r".*"}
        self.groups = {k: fr"{v}" for k, v in groups.items()}

        if subgroups is None:
            subgroups = {"all": r".*"}
        self.subgroups = {k: fr"{v}" for k, v in subgroups.items()}

        self.add_legend = add_legend
        self.write_chemiscope = write_chemiscope

        return
    
    def forward(self, dataset, describer):
        """"""
        super().forward()

        describer.directory = self.directory
        features = describer.run(dataset=dataset)

        starts = [0] + [len(d._images) for d in dataset]
        starts = np.cumsum(starts)
        self._debug(f"starts: {starts}")

        # - get groups
        group_indices = {k: {sk: [] for sk in self.subgroups} for k in self.groups}
        for i, system in enumerate(dataset):
            # -- match group
            for k, v in self.groups.items():
                if re.match(v, system.prefix) is not None:
                    self._print(f"{v}, {system.prefix}")
                    break
            else:
                continue
            # -- match subgroups
            for sk, sv in self.subgroups.items():
                curr_indices = [x+starts[i] for x in system.get_matched_indices(sv)]
                #if sk not in group_indices[k]:
                #    group_indices[k][sk] = curr_indices
                #else:
                #    group_indices[k][sk].extend(curr_indices)
                group_indices[k][sk].extend(curr_indices)
        #self._debug(f"groups: {group_indices}")

        # - merge groups
        merged_groups = {}
        for name, curr_group in group_indices.items():
            for subname, indices in curr_group.items():
                if self.level == 0:
                    gname = f"{name}+{subname}"
                elif self.level == 1:
                    gname = f"{name}"
                elif self.level == 2:
                    gname = f"{subname}"
                else:
                    raise RuntimeError()
                if gname in merged_groups:
                    merged_groups[gname].extend(indices)
                else:
                    merged_groups[gname] = indices

        self._plot_results(features, merged_groups, self.add_legend)

        # - save chemiscope?
        if self.write_chemiscope:
            frames = []
            for d in dataset:
                frames.extend(d._images)
            self._write_chemiscope(frames=frames, features=features)

        return

    def _plot_results(self, features, groups: Mapping[str,List[int]], add_legend, *args, **kwargs):
        """"""
        # - plot selection
        import matplotlib.pyplot as plt
        try:
            plt.style.use("presentation")
        except Exception as e:
            ...

        from sklearn.decomposition import PCA

        reducer = PCA(n_components=2)
        reducer.fit(features)

        # -- separate
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))

        curves = []
        for i, (name, indices) in enumerate(groups.items()):
            print(f"{name} -> {len(indices)}")
            if len(indices) > 0:
                proj = reducer.transform(features[indices,:])
                curve = ax.scatter(
                    proj[:, 0], proj[:, 1], alpha=0.25, zorder=100-i,
                    label=f"{name} {len(indices)}"
                )
                curves.append(curve)
        labels = [c.get_label() for c in curves]

        ax.axis("off")
        fig.savefig(self.directory/"pca.png", transparent=True)

        label_params = ax.get_legend_handles_labels()
        if add_legend:
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
            ax.axis("off")
            ax.legend(*label_params)
            fig.savefig(self.directory/"pca-legend.png", transparent=True)

        return 

    def _write_chemiscope(self, frames, features):
        """"""
        USE_CHEMISCOPE = 0
        try:
            import chemiscope
            USE_CHEMISCOPE = 1
        except Exception as e:
            print(e)

        from sklearn.decomposition import PCA
        # - write chemiscope inputs
        pca = PCA(n_components=2).fit_transform(features)
        properties = dict(
            PCA = dict(
                target = "structure",
                values = pca
            ),
            #energies = dict(
            #    target = "structure",
            #    values = [a.get_potential_energy() for a in frames],
            #    units = "eV"
            #)
        )

        frame_properties = chemiscope.extract_properties(
            frames,
            only=["energy"]
        )
        properties.update(**frame_properties)

        chemiscope.write_input(
            str(self.directory/"chemiscope.json.gz"), 
            frames=frames, properties=properties
        )

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import uuid
import pathlib
import shutil
import tarfile
import time
from typing import Tuple, List, NoReturn, Union
import tempfile
import warnings
import json
import yaml

import numpy as np

from tinydb import Query, TinyDB

from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write

from .. import config
from ..builder.builder import StructureBuilder
from ..computation.driver import AbstractDriver
from ..potential.manager import AbstractPotentialManager
from ..utils.command import CustomTimer

from .worker import AbstractWorker
from .utils import copy_minimal_frames, get_file_md5


"""Monitor computation tasks with Worker.

"""


def compare_atoms(a1, a2):
    """Compare structures according to cell, chemical symbols, and positions.

    Structures will be different if the atom order changes.

    """
    c1 = a1.get_cell(complete=True)
    c2 = a2.get_cell(complete=True)
    if np.sum(c1 - c2) >= 1e-8:
        return False

    s1 = a1.get_chemical_symbols()
    s2 = a2.get_chemical_symbols()
    if s1 != s2:
        return False

    p1 = a1.get_positions()
    p2 = a2.get_positions()
    if np.sum(p1 - p2) >= 1e-8:
        return False

    return True


class DriverBasedWorker(AbstractWorker):
    """Monitor driver-based jobs.

    Lifetime: queued (running) -> finished -> retrieved

    Note:
        The database stores each unique job ID and its working directory.

    """

    #: How many structures performed in one job.
    batchsize: int = 1

    #: Print period of share_wdir calculations.
    print_period: int = 100

    #: Reserved keys in atoms.info by gdp.
    reserved_keys: List["str"] = ["energy", "step", "wdir"]

    #: Attached driver object.
    _driver = None

    #: Whether generate an independant random_seed for each candidate's driver.
    _share_random_seed: bool = False

    #: Whether share calc dir for each candidate. Only for command run and spc.
    _share_wdir: bool = False

    #: Whether retain the info stored in atoms and add to trajectory.
    _retain_info: bool = False

    def __init__(
        self, potter_, driver_=None, scheduler_=None, directory_=None, *args, **kwargs
    ):
        """"""
        self.batchsize = kwargs.pop("batchsize", 1)

        assert isinstance(potter_, AbstractPotentialManager), ""

        self.potter = potter_
        self.driver = driver_
        self.scheduler = scheduler_
        if directory_:
            self.directory = directory_

        self.n_jobs = config.NJOBS

        return

    @property
    def driver(self) -> AbstractDriver:
        return self._driver

    @driver.setter
    def driver(self, driver_):
        """"""
        assert isinstance(driver_, AbstractDriver), ""
        # TODO: check driver is consistent with potter
        self._driver = driver_
        return

    def _split_groups(
        self, nframes: int, batchsize: int = 1
    ) -> Tuple[List[int], List[int]]:
        """Split nframes into groups."""
        # - split frames
        self._debug(f"split_groups for {nframes} nframes and {batchsize} batchsize.")
        ngroups = int(np.floor(1.0 * nframes / batchsize))
        group_indices = [0]
        for i in range(ngroups):
            group_indices.append((i + 1) * batchsize)
        if group_indices[-1] != nframes:
            group_indices.append(nframes)
        starts, ends = group_indices[:-1], group_indices[1:]
        assert len(starts) == len(ends), "Inconsistent start and end indices..."
        # group_indices = [f"{s}:{e}" for s, e in zip(starts,ends)]

        return (starts, ends)

    def _read_cached_info(self):
        # - read extra info data
        _info_data = []
        for p in (self.directory / "_data").glob("*_info.txt"):
            identifier = p.name[: self.UUIDLEN]  # MD5
            with open(p, "r") as fopen:
                for line in fopen.readlines():
                    if not line.startswith("#"):
                        _info_data.append(line.strip().split())
        _info_data = sorted(_info_data, key=lambda x: int(x[0]))

        return _info_data

    def _read_cached_xinfo(self):
        # - read extra info data
        info_keys, _info_data = [], []
        for p in (self.directory / "_data").glob("*_xinfo.txt"):
            identifier = p.name[: self.UUIDLEN]  # MD5
            with open(p, "r") as fopen:
                lines = fopen.readlines()
                info_keys = lines[0].split()[1:]
                for line in lines:
                    if not line.startswith("#"):
                        _info_data.append(line.strip().split()[1:])
        # _info_data = sorted(_info_data, key=lambda x: int(x[0]))
        assert info_keys, f"info_keys must not be empty."

        return info_keys, _info_data

    def _preprocess(self, builder, *args, **kwargs):
        """"""
        # - get frames
        frames = []
        if isinstance(builder, StructureBuilder):
            frames = builder.run()
        else:
            assert all(
                isinstance(x, Atoms) for x in frames
            ), "Input should be a list of atoms."
            frames = builder
        prev_frames = frames

        # - check differences of input structures
        processed_dpath = self.directory / "_data"
        processed_dpath.mkdir(exist_ok=True)

        # - NOTE: atoms.info is a dict that does not maintain order
        #         thus, the saved file maybe different
        curr_frames, curr_info = copy_minimal_frames(prev_frames)

        # -- get MD5 of current input structures
        # NOTE: if two jobs start too close,
        #       there may be conflicts in checking structures
        with tempfile.NamedTemporaryFile(mode="w", suffix=".xyz") as tmp:
            write(tmp.name, curr_frames, columns=["symbols", "positions", "move_mask"])

            with open(tmp.name, "rb") as fopen:
                curr_md5 = get_file_md5(fopen)

        # - generate random_seeds
        #   NOTE: no matter the job status is, they are generated to sync rng state
        self._print(f"Driver's random_seed: {self.driver.random_seed}")
        if not self._share_random_seed:
            rng = np.random.Generator(np.random.PCG64(self.driver.random_seed))
            random_seeds = rng.integers(0, 1e8, size=len(curr_frames))
            random_seeds = [int(x) for x in random_seeds]  # We need List[int]!!
        else:
            random_seeds = [self.driver.random_seed] * len(curr_frames)

        # - check info data
        _info_data = self._read_cached_info()

        stored_fname = f"{curr_md5}.xyz"
        if (processed_dpath / stored_fname).exists():
            self._print(f"Found file with md5 {curr_md5}")
            self._info_data = _info_data
            start_confid = 0
            for x in self._info_data:
                if x[1] == curr_md5:
                    break
                start_confid += 1
            # NOTE: compatability
            if len(_info_data[0]) > 5:
                random_seeds = [int(x[-1]) for x in _info_data]
            else:
                ...  # Use generated random seeds
        else:
            if self._retain_info:
                info_keys = []
                for a in prev_frames:
                    info_keys.extend(list(a.info.keys()))
                info_keys = sorted(set(info_keys))
                content = (
                    f'{"#id":<12s}  '
                    + ("{:<24s}  " * len(info_keys)).format(*info_keys)
                    + "\n"
                )
                for i, a in enumerate(prev_frames):
                    line = (
                        f"{i:<24d}  "
                        + "  ".join([f"{str(a.info.get(k)):<24s}" for k in info_keys])
                        + "\n"
                    )
                    content += line
                with open(processed_dpath / f"{curr_md5}_xinfo.txt", "w") as fopen:
                    fopen.write(content)
            # - save structures
            write(
                processed_dpath / stored_fname,
                curr_frames,
                # columns=["symbols", "positions", "momenta", "tags", "move_mask"]
            )
            # - save current atoms.info and append curr_info to _info_data
            start_confid = len(_info_data)
            content = "{:<12s}  {:<32s}  {:<12s}  {:<12s}  {:<s}  {:>24s}\n".format(
                "#id", "MD5", "confid", "step", "wdir", "rs"
            )
            for i, ((confid, step, wdir), rs) in enumerate(
                zip(curr_info, random_seeds)
            ):
                line = "{:<12d}  {:<32s}  {:<12d}  {:<12d}  {:<s}  {:>24d}\n".format(
                    i + start_confid, curr_md5, confid, step, wdir, rs
                )
                content += line
                _info_data.append(line.strip().split())
            self._info_data = _info_data
            with open(processed_dpath / f"{curr_md5}_info.txt", "w") as fopen:
                fopen.write(content)

        return curr_md5, curr_frames, start_confid, random_seeds

    def _prepare_batches(
        self, frames: List[Atoms], start_confid: int, rng_states: Union[List[int], List[dict]]
    ):
        # - check wdir
        num_frames = len(frames)
        # NOTE: get a list even if it only has one structure
        # TODO: a better strategy to deal with wdirs...
        #       conflicts:
        #           merged trajectories from different drivers that all have cand0
        wdirs = []  # [(confid,dynstep), ..., ()]
        for i, atoms in enumerate(frames):
            # -- set wdir
            wdir = "cand{}".format(int(self._info_data[i + start_confid][0]))
            wdirs.append(wdir)
            atoms.info["wdir"] = wdir
        # - check whether each structure has a unique wdir
        assert (
            len(set(wdirs)) == num_frames
        ), f"Found duplicated wdirs {len(set(wdirs))} vs. {num_frames}..."

        # - split structures into different batches
        if self._share_wdir:
           self._print(f"Worker overwrites batchsize to {num_frames =} as it uses share_wdir.")
           batchsize = num_frames
        else:
           batchsize = self.batchsize

        starts, ends = self._split_groups(num_frames, batchsize)

        batches = []
        for i, (s, e) in enumerate(zip(starts, ends)):
            # - prepare structures and dirnames
            global_indices = range(s, e)
            # NOTE: get a list even if it only has one structure
            cur_frames = [frames[x] for x in global_indices]
            cur_wdirs = [wdirs[x] for x in global_indices]
            curr_rs = [rng_states[x] for x in global_indices]
            for x in cur_frames:
                x.info["group"] = i
            # - check whether each structure has a unique wdir
            assert len(set(cur_wdirs)) == len(
                cur_frames
            ), f"Found duplicated wdirs {len(set(wdirs))} vs. {len(cur_frames)} for group {i}..."

            # - set specific params
            batches.append([global_indices, cur_wdirs, curr_rs])

        return batches

    def run(self, builder=None, rng_states=list(), *args, **kwargs) -> None:
        """Split frames into groups and submit jobs."""
        super().run(*args, **kwargs)

        # - check if the same input structures are provided
        identifier, frames, start_confid, new_rng_states = self._preprocess(builder)
        if rng_states: # Sometimes we need explicit rng_states as in active learning
            new_rng_states = rng_states
        batches = self._prepare_batches(frames, start_confid, new_rng_states)

        # - read metadata from file or database
        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            queued_jobs = database.search(Query().queued.exists())
        queued_names = [q["gdir"][self.UUIDLEN + 1 :] for q in queued_jobs]
        queued_frames = [q["md5"] for q in queued_jobs]

        for ig, (global_indices, wdirs, rs) in enumerate(batches):
            # - set job name
            batch_name = f"group-{ig}"
            uid = str(uuid.uuid1())
            job_name = uid + "-" + batch_name

            # -- whether store job info
            if self.scheduler.name != "local":
                if batch_name in queued_names and identifier in queued_frames:
                    self._print(f"{batch_name} at {self.directory.name} was submitted.")
                    continue
            else:  # Local Scheduler
                is_resubmit = kwargs.get("resubmit", False)
                if not is_resubmit:
                    # NOTE:  Only re-run computation when resubmit is set
                    if batch_name in queued_names and identifier in queued_frames:
                        self._print(
                            f"{batch_name} at {self.directory.name} was submitted."
                        )
                        continue
                else:
                    ...

            # - specify which group this worker is responsible for
            #   if not, then skip
            #   Skip batch here assures the skipped batches will not recorded and
            #   thus will not affect their execution if several batches run at the same time.
            target_number = kwargs.get("batch", None)
            if isinstance(target_number, int):
                if ig != target_number:
                    with CustomTimer(name="run-driver", func=self._print):
                        self._print(
                            f"{time.asctime( time.localtime(time.time()) )} {self.driver.directory.name} batch {ig} is skipped..."
                        )
                        continue
                else:
                    ...
            else:
                ...

            # - run batch
            # NOTE: For command execution, if computation exits incorrectly,
            #       it will not be recorded. The computation will resume next
            #       time.
            self._irun(
                batch_name,
                uid,
                identifier,
                frames,
                global_indices,
                wdirs,
                rng_states=rs,
                *args,
                **kwargs,
            )

            # - save this batch job to the database
            if identifier not in queued_frames:
                with TinyDB(
                    self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
                ) as database:
                    _ = database.insert(
                        dict(
                            uid=uid,
                            md5=identifier,
                            gdir=job_name,
                            group_number=ig,
                            wdir_names=wdirs,
                            queued=True,
                        )
                    )
                # save worker input for later review
                worker_input_fpath = self.directory/"_data"/f"worker-{identifier}.json"
                if not worker_input_fpath.exists():
                    with open(worker_input_fpath, "w") as fopen:
                        json.dump(self.as_dict(), fopen, indent=2)

        return

    def _irun(self, *args, **kwargs):
        """"""

        raise NotImplementedError("Function to run a batch of structures is undefined.")

    def inspect(self, resubmit=False, *args, **kwargs):
        """Check if any job were finished correctly not due to time limit.

        Args:
            resubmit: Check whether submit unfinished jobs.

        """
        self._initialise(*args, **kwargs)
        self._print(f"~~~{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs()

        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in running_jobs:
                self._debug(f"inspect {job_name}")
                doc_data = database.get(Query().gdir == job_name)
                uid = doc_data["uid"]
                identifier = doc_data["md5"]
                batch = doc_data["group_number"]

                # self.scheduler.set(**{"job-name": job_name})
                self.scheduler.job_name = job_name
                self.scheduler.script = self.directory / f"run-{uid}.script"

                # -- check whether the jobs if running
                if self.scheduler.is_finished():  # if it is still in the queue
                    # -- valid if the task finished correctly not due to time-limit
                    is_finished = False
                    wdir_names = doc_data["wdir_names"]
                    if not self._share_wdir:
                        # - first a quick check if all wdirs exist
                        wdir_existence = [
                            (self.directory / x).exists() for x in wdir_names
                        ]
                        nwdir_exists = sum(1 for x in wdir_existence if x)
                        if all(wdir_existence):
                            for x in wdir_names:
                                curr_wdir = self.directory / x
                                self.driver.directory = curr_wdir
                                if not self.driver.read_convergence():
                                    self._print(
                                        f"Found unfinished computation at {curr_wdir.name}"
                                    )
                                    break
                            else:
                                is_finished = True
                        else:
                            self._print("NOT ALL wdirs exist.")
                        self._print(f"progress: {nwdir_exists}/{len(wdir_existence)}")
                    else:
                        cache_frames = read(
                            self.directory / "_data" / f"{identifier}_cache.xyz", ":"
                        )
                        cache_wdirs = [a.info["wdir"] for a in cache_frames]
                        if set(wdir_names) == set(cache_wdirs):
                            is_finished = True
                            self._print(
                                f"Found unfinished computation at cand{len(cache_wdirs)}"
                            )
                    if is_finished:
                        # -- finished correctly
                        self._print(f"{job_name} is finished...")
                        doc_data = database.get(Query().gdir == job_name)
                        database.update({"finished": True}, doc_ids=[doc_data.doc_id])
                    else:
                        # NOTE: no need to remove unfinished structures
                        #       since the driver would check it
                        # BUG: If batchsize == 1, the resbumit run many times
                        #      This is not as expected.
                        if resubmit:
                            if self.scheduler.name != "local":
                                jobid = self.scheduler.submit()
                                self._print(
                                    f"{job_name} is re-submitted with JOBID {jobid}."
                                )
                            else:
                                frames = read(
                                    self.directory / "_data" / f"{identifier}.xyz", ":"
                                )
                                self.run(frames, batch=batch, resubmit=True)
                else:
                    self._print(f"{job_name} is running...")

        return

    def retrieve(
        self,
        include_retrieved: bool = False,
        given_wdirs: List[str] = None,
        use_archive: bool = False,
        *args,
        **kwargs,
    ):
        """Read results from wdirs.

        Args:
            include_retrieved: Whether include wdirs that are already retrieved.
                              Otherwise, all finished jobs are included.

        Returns:
            A nested List of Atoms.

        """
        self.inspect(*args, **kwargs)
        self._print(f"~~~{self.__class__.__name__}+retrieve")

        # NOTE: sometimes retrieve is used without run
        self._info_data = self._read_cached_info()  # update _info_data

        # - check status and get latest results
        unretrieved_wdirs_ = []
        if not include_retrieved:
            unretrieved_jobs = self._get_unretrieved_jobs()
        else:
            unretrieved_jobs = self._get_finished_jobs()

        unretrieved_identifiers = []

        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                unretrieved_identifiers.append(doc_data["md5"])
                unretrieved_wdirs_.extend(
                    self.directory / w for w in doc_data["wdir_names"]
                )

        # - get given wdirs
        unretrieved_wdirs = []
        if given_wdirs is not None:
            for wdir in unretrieved_wdirs_:
                wdir_name = wdir.name
                if wdir_name in given_wdirs:
                    unretrieved_wdirs.append(wdir)
        else:
            unretrieved_wdirs = unretrieved_wdirs_

        # - read results
        if unretrieved_wdirs:
            # - read results
            unretrieved_wdirs = [pathlib.Path(x) for x in unretrieved_wdirs]
            if not self._share_wdir:
                archive_path = (self.directory / "cand.tgz").absolute()
                if not archive_path.exists():  # read unarchived data
                    results = self._read_results(unretrieved_wdirs, *args, **kwargs)
                else:
                    self._print("read archived data...")
                    results = self._read_results(
                        unretrieved_wdirs, archive_path=archive_path, *args, **kwargs
                    )
                # - archive results if it has not been done
                if use_archive and not archive_path.exists():
                    self._print("archive computation folders...")
                    with tarfile.open(archive_path, "w:gz", compresslevel=6) as tar:
                        for w in unretrieved_wdirs:
                            tar.add(w, arcname=w.name)
                    for w in unretrieved_wdirs:
                        shutil.rmtree(w)
                else:
                    ...
            else:
                # TODO: deal with traj...
                cache_frames = []
                for identifier in unretrieved_identifiers:
                    cache_frames.extend(
                        read(self.directory / "_data" / f"{identifier}_cache.xyz", ":")
                    )
                wdir_names = [x.name for x in unretrieved_wdirs]
                results_ = [a for a in cache_frames if a.info["wdir"] in wdir_names]
                # - convert to a List[List[Atoms]] as non-shared run
                results_ = [[a] for a in results_]
                # -- re-add info
                if self._retain_info:
                    info_keys, info_data = self._read_cached_xinfo()
                    retained_keys = [
                        k for k in info_keys if k not in self.reserved_keys
                    ]
                    for i, traj_frames in enumerate(results_):
                        retained_dict = {
                            k: v
                            for k, v in zip(info_keys, info_data[i])
                            if k in retained_keys and v is not None
                        }
                        traj_frames[0].info.update(retained_dict)
                results = results_
        else:
            results = []

        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                database.update({"retrieved": True}, doc_ids=[doc_data.doc_id])

        return results

    def _read_results(
        self,
        unretrieved_wdirs: List[pathlib.Path],
        archive_path: pathlib.Path = None,
        *args,
        **kwargs,
    ) -> Union[List[Atoms], List[List[Atoms]]]:
        """Read results from calculation directories.

        Args:
            unretrieved_wdirs: Calculation directories.

        """
        with CustomTimer(name="read-results", func=self._print):
            # NOTE: works for vasp, ...
            results_ = Parallel(n_jobs=self.n_jobs)(
                delayed(self._iread_results)(
                    self.driver,
                    wdir,
                    info_data=self._info_data,
                    archive_path=archive_path,
                )
                for wdir in unretrieved_wdirs
            )

            # -- re-add info
            if self._retain_info:
                info_keys, info_data = self._read_cached_xinfo()
                retained_keys = [k for k in info_keys if k not in self.reserved_keys]
                for i, traj_frames in enumerate(results_):
                    retained_dict = {
                        k: v
                        for k, v in zip(info_keys, info_data[i])
                        if k in retained_keys and v is not None
                    }
                    traj_frames[0].info.update(retained_dict)

            # NOTE: Failed Calcution, One fail, traj fails
            results = []
            for i, traj_frames in enumerate(results_):
                # - sift error structures
                if traj_frames:
                    results.append(traj_frames)
                else:
                    self._print(
                        f"Found empty calculation at {str(self.directory)} with cand{i}..."
                    )

            if results:
                self._print(
                    f"new_trajectories: {len(results)} nframes of the first: {len(results[0])}"
                )

        return results

    @staticmethod
    def _iread_results(
        driver, wdir, info_data: dict = None, archive_path: pathlib.Path = None
    ) -> List[Atoms]:
        """Extract results from a single directory.

        This must be a staticmethod as it may be pickled by joblib for parallel
        running.

        Args:
            wdir: Working directory.

        """
        driver.directory = wdir
        # NOTE: name convention, cand1112_field1112_field1112
        confid_ = int(wdir.name.strip("cand").split("_")[0])  # internal name
        if info_data is not None:
            cache_confid = int(info_data[confid_][2])
            if cache_confid >= 0:
                confid = cache_confid
            else:
                confid = confid_
        else:
            confid = confid_

        # NOTE: always return the entire trajectories
        traj_frames = driver.read_trajectory(
            add_step_info=True, archive_path=archive_path
        )
        for a in traj_frames:
            a.info["confid"] = confid
            a.info["wdir"] = str(wdir.name)

        return traj_frames
    
    def _write_worker_inputs(self, uid: str):
        """"""
        worker_params = {}
        worker_params["driver"] = self.driver.as_dict()
        worker_params["potential"] = self.potter.as_dict()
        worker_params["batchsize"] = self.batchsize
        worker_params["share_wdir"] = self._share_wdir
        worker_params["retain_info"] = self._retain_info

        with open(self.directory / f"worker-{uid}.yaml", "w") as fopen:
            yaml.dump(worker_params, fopen)

        return

    def as_dict(self) -> dict:
        """"""
        worker_params = super().as_dict()
        worker_params["batchsize"] = self.batchsize

        return worker_params


class QueueDriverBasedWorker(DriverBasedWorker):

    def _irun(
        self,
        batch_name: str,
        uid: str,
        identifier: str,
        frames: List[Atoms],
        curr_indices: List[int],
        curr_wdirs: List[Union[str, pathlib.Path]],
        rng_states: Union[List[int], List[dict]],
        *args,
        **kwargs,
    ) -> None:
        """"""
        batch_number = int(batch_name.split("-")[-1])

        # - save worker file
        self._write_worker_inputs(uid=uid)
        
        # - save structures
        dataset_path = str((self.directory / "_data" / f"{identifier}.xyz").resolve())

        # - save scheduler file
        jobscript_fname = f"run-{uid}.script"
        self.scheduler.job_name = uid + "-" + batch_name
        self.scheduler.script = self.directory / jobscript_fname

        self.scheduler.user_commands = "gdp -p {} compute {} --batch {} --spawn\n".format(
            (self.directory / f"worker-{uid}.yaml").name,
            # (self.directory/structure_fname).name
            dataset_path,
            batch_number,
        )

        # - TODO: check whether params for scheduler is changed
        self.scheduler.write()
        if self._submit:
            self._print(f"{self.directory.name} JOBID: {self.scheduler.submit()}")
        else:
            self._print(f"{self.directory.name} waits to submit.")

        return


class CommandDriverBasedWorker(DriverBasedWorker):

    def _irun(
        self,
        batch_name: str,
        uid: str,
        identifier: str,
        frames: List[Atoms],
        curr_indices: List[int],
        curr_wdirs: List[Union[str, pathlib.Path]],
        rng_states: Union[List[int], List[dict]],
        *args,
        **kwargs,
    ) -> None:
        """Run calculations directly in the command line.

        Local execution supports a compact mode as structures will reuse the
        calculation working directory.

        """
        batch_number = int(batch_name.split("-")[-1])

        # - get structures
        curr_frames = [frames[i] for i in curr_indices]

        # - run calculations
        with CustomTimer(name="run-driver", func=self._print):
            if not self._share_wdir:
                for wdir, atoms, rs in zip(curr_wdirs, curr_frames, rng_states):
                    self.driver.directory = self.directory / wdir
                    prev_random_seed = self.driver.random_seed
                    self.driver.set_rng(seed=rs)
                    self._print(
                        f"{time.asctime( time.localtime(time.time()) )} {str(wdir)} {self.driver.directory.name} is running..."
                    )
                    self.driver.reset()
                    self.driver.run(atoms, read_ckpt=True, extra_info=None)
                    self.driver.set_rng(seed=prev_random_seed)
            else:
                # read calculation cache
                cache_fpath = self.directory / "_data" / f"{identifier}_cache.xyz"
                if cache_fpath.exists():
                    cache_frames = read(cache_fpath, ":")
                    cache_wdirs = [a.info["wdir"] for a in cache_frames]
                else:
                    cache_wdirs = []

                num_curr_frames = len(curr_frames)

                # run calculations
                temp_wdir = self.directory / "_shared"
                for i, (wdir, atoms, rs) in enumerate(zip(curr_wdirs, curr_frames, rng_states)):
                    if wdir in cache_wdirs:
                        continue
                    if temp_wdir.exists():
                        shutil.rmtree(temp_wdir)
                    self.driver.directory = temp_wdir
                    if i % self.print_period == 0 or i+1 == num_curr_frames:
                        self._print(
                            f"{time.asctime( time.localtime(time.time()) )} {str(wdir)} {self.driver.directory.name} is running..."
                        )
                    self.driver.set_rng(seed=rs)
                    self.driver.reset()
                    new_atoms = self.driver.run(
                        atoms, read_ckpt=False, extra_info=dict(wdir=wdir)
                    )
                    # - save data
                    # TODO: There may have conflicts in write as many groups may run at the same time.
                    #       Add protection to the file.
                    write(
                        self.directory / "_data" / f"{identifier}_cache.xyz",
                        new_atoms,
                        append=True,
                    )

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

from ..core.register import Register

# - driver (dynamics) backends...
register_drivers = Register("driver")

from gdpx.computation.asedriver import AseDriver
register_drivers.register("ase")(AseDriver)

from gdpx.computation.lammps import LmpDriver
register_drivers.register("lammps")(LmpDriver)

from gdpx.computation.lasp import LaspDriver
register_drivers.register("lasp")(LaspDriver)

from gdpx.computation.vasp import VaspDriver
register_drivers.register("vasp")(VaspDriver)

from gdpx.computation.cp2k import Cp2kDriver
register_drivers.register("cp2k")(Cp2kDriver)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List
from pathlib import Path

import numpy as np

from ase import Atoms
from ase.constraints import FixAtoms

from .builder import StructureBuilder


class DimerBuilder(StructureBuilder):

    def __init__(self, elements: List[str], distances: List[float]=[0.8,2.5,0.05], directory=Path.cwd(), *args, **kwargs):
        """"""
        super().__init__(directory, *args, **kwargs)

        self.elements = elements
        assert len(self.elements) == 2, "DimerBuilder needs two chemical symbols as elements."

        self.distances = distances
        assert len(self.distances) == 3, "DimerBuilder needs min, max and intv for the distance."

        return

    def run(self, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(*args, **kwargs)

        dmin, dmax, intv = self.distances
        distances = np.arange(dmin, dmax+intv, intv)

        frames = []
        for dis in distances:
            atoms = Atoms(
                symbols=self.elements, 
                positions=[
                    [0., 0., 0.],
                    [0., 0., dis]
                ],
                #cell = 20.*np.eye(3),
                cell = [[19., 0., 0.], [0., 20., 0.], [0., 0., 21.]],
                pbc=[True,True,True]
            )
            atoms.set_constraint(FixAtoms(indices=[0]))
            frames.append(atoms)

        return frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools
import functools
import operator
import numbers
import pathlib
from typing import Any, Optional, List, Mapping

import h5py
import numpy as np

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

#:
GROUP_NAME = "array2d"

#:
RETAINED_INFO_NAMES: List[str] = [
    "confid", "step",
    "max_devi_e", "min_devi_e", "avg_devi_e",
    "max_devi_v", "min_devi_v", "avg_devi_v",
    "max_devi_f", "min_devi_f", "avg_devi_f"
]

#:
RETAIEND_INFO_DTYPES: List[str] = [
    "i8", "i8",
    "f", "f", "f",
    "f", "f", "f",
    "f", "f", "f"
]

#: Saved calculated property names.
RETAINED_CALC_PROPS: List[str] = ["energy", "free_energy", "forces"]

#: Saved calculated atomic property names.
RETAINED_ATOMIC_CALC_PROPS: List[str] = ["forces"]


def _flat_data(items):
    """"""
    # if not isinstance(ret, Atoms):
    if isinstance(items, list) and not isinstance(items[0], Atoms):
        items = _flat_data(list(itertools.chain(*items)))

    return items


def _reshape_data(data, shape):
    """"""
    data = data  # NOTE: A PURE LIST
    for i, tsize in enumerate(shape[::-1][:-1]):
        npoints = len(data)
        length = int(npoints/tsize)
        data_ = []
        for i in range(length):
            data_.append(data[i*tsize:(i+1)*tsize])
        data = data_

    return data


def _map_idx(loc, shape):
    """"""
    i = 0
    for dim, j in enumerate(loc):
        i += j*functools.reduce(operator.mul, ([1]+list(shape[dim+1:])))
    return i


class AtomsNDArray:

    #: Atoms data.
    _data: Optional[List[Atoms]] = None

    #: Array shape.
    _shape: Optional[List[int]] = None

    #: Array-like indices.
    _markers = None

    #:
    _ind_map: Optional[Mapping[int, int]] = None

    """Define an atoms array object.

    This definition gives correct interval selection.

    """

    def __init__(self, data: list = None, markers=None) -> None:
        """Init from a List^n object."""
        # TODO: Check data should be a list
        if data is None:
            data = []
        # TODO: Check data should be a list
        if isinstance(data, list):
            data = data
        elif isinstance(data, AtomsNDArray):
            data = data.tolist()
        else:
            raise ValueError(
                f"Data should be a list or a AtomsNDArray instead of {type(data)}.")

        if len(data) == 0:
            raise RuntimeError(f"Input data is empty as {data}.")

        self._shape, self._data, self._markers, self._ind_map = self._process_data(
            data)

        # TODO: Check IndexError?
        if markers is not None:
            self.markers = markers

        return

    @staticmethod
    def _process_data(data_nd):
        """"""
        sizes = [[len(data_nd)]]

        def _flat_inhomo_data(items: list):
            """"""
            if isinstance(items, list) or isinstance(items, tuple):
                # NOTE: The input items must be a nested list only with Atoms or None elements
                #print(f"current size: {sizes}")
                #print(f"current size: {items[0]}")
                if not (isinstance(items[0], Atoms) or items[0] is None):
                    sizes.append([len(item) for item in items])
                    items = _flat_inhomo_data(list(itertools.chain(*items)))
                else:
                    return items
            else:
                ...

            return items

        # NOTE: properly deal with None?
        data_1d = [a for a in _flat_inhomo_data(data_nd) if a is not None]
        shape = tuple([max(s) for s in sizes])
        # print(f"sizes: {sizes}")
        # print(f"shape: {shape}")

        def assign_markers(arr, seq):
            # print(arr)
            # print(seq)
            if isinstance(arr, list):  # assume it is a list
                if arr[0] is None:  # arr.ndim == 1:
                    inds = []
                    for i, a in enumerate(seq):
                        if isinstance(a, Atoms):
                            inds.append(i)
                    m = len(arr)
                    for i in range(m):
                        if i in inds:
                            arr[i] = True
                        else:
                            arr[i] = False
                else:
                    for subarr, subseq in itertools.zip_longest(arr, seq, fillvalue=()):
                        assign_markers(subarr, subseq)
            else:
                ...

        raw_markers = np.full(shape, None).tolist()
        _ = assign_markers(raw_markers, data_nd)
        markers_1d = np.argwhere(raw_markers)
        ind_map = {_map_idx(loc, shape): i for i, loc in enumerate(markers_1d)}

        return shape, data_1d, markers_1d, ind_map

    @property
    def shape(self):
        """"""

        return self._shape

    @property
    def ndim(self) -> int:
        """"""

        return len(self.shape)

    @property
    def raw_markers(self):
        """"""
        raw_markers = np.full(self.shape, None)
        for m in self.markers:
            raw_markers[tuple(m)] = True

        return raw_markers

    @property
    def markers(self):
        """Return markers.

        If it is the first time, all structures are returned.

        """

        return self._markers

    @markers.setter
    def markers(self, new_markers):
        """Set new markers.

        Args:
            new_markers: These should have the shape as the array.

        """
        # TODO: IndexError?
        if isinstance(new_markers, list):
            ...
        elif isinstance(new_markers, np.ndarray):
            new_markers = new_markers.tolist()
        else:
            raise ValueError("Index must be a list or a ndarray.")

        self._markers = np.array(sorted(new_markers))

        return

    def get_marked_structures(self, markers=None):
        """Get structures according to markers.

        If custom markers is None, `self._markers` will be used instead.

        """
        if markers is None:
            curr_markers = self.markers
        else:
            curr_markers = markers

        structures = [self._data[self._ind_map[_map_idx(
            loc, self.shape)]] for loc in curr_markers]

        return structures

    def tolist(self):
        """"""
        data_1d = np.full(self.shape, None).flatten().tolist()
        for k, v in self._ind_map.items():
            data_1d[k] = self._data[v]

        return _reshape_data(data_1d, self.shape)

    @classmethod
    def from_file(cls, target):
        """"""
        with h5py.File(target, "r") as fopen:
            grp = fopen.require_group("images")
            shape = grp.attrs["shape"]
            images = cls._from_hd5grp(grp=grp)
            markers = np.array(grp["markers"][:])
            mapper = {k: v for k, v in zip(grp["map_k"], grp["map_v"])}

        data_1d = np.full(shape, None).flatten().tolist()
        for k, v in mapper.items():
            data_1d[k] = images[v]

        data = _reshape_data(data_1d, shape=shape)

        return cls(data=data, markers=markers)

    @classmethod
    def _from_hd5grp(cls, grp):
        """Reconstruct an atoms_array from data stored in HDF5 group `images`."""
        # - rebuild structures
        natoms_list = grp["natoms"]

        images = []
        for natoms, box, pbc, atomic_numbers, positions in zip(
            natoms_list, grp["box"], grp["pbc"], grp["atype"], grp["positions"]
        ):
            atoms = Atoms(
                numbers=atomic_numbers[:natoms], positions=positions[:natoms, :],
                cell=box.reshape(3, 3), pbc=pbc
            )
            images.append(atoms)
        nimages = len(images)

        # -- add info
        for name in RETAINED_INFO_NAMES:
            data = grp.get(name, default=None)
            if data is not None:
                for atoms, v in zip(images, data):
                    atoms.info[name] = v

        # -- add calc
        results = [{} for _ in range(nimages)]  # List[Mapping[str,data]]
        for name in RETAINED_CALC_PROPS:
            data = grp.get(name, default=None)
            if data is not None:
                for i, v in enumerate(data):
                    if name not in RETAINED_ATOMIC_CALC_PROPS:
                        results[i][name] = v
                    else:
                        results[i][name] = v[:natoms_list[i]]

        for atoms, ret in zip(images, results):
            spc = SinglePointCalculator(atoms, **ret)
            atoms.calc = spc

        return images

    def save_file(self, target):
        """"""
        with h5py.File(target, mode="w") as fopen:
            grp = fopen.create_group("images")
            grp.attrs["shape"] = self.shape

            # - save structures
            self._convert_images(grp=grp, images=self._data)

            # - save markers
            grp.create_dataset("markers", data=self.markers, dtype="i8")

            # - save mapper
            mapper_k, mapper_v = [], []
            for k, v in self._ind_map.items():
                mapper_k.append(k)
                mapper_v.append(v)
            grp.create_dataset("map_k", data=mapper_k, dtype="i8")
            grp.create_dataset("map_v", data=mapper_v, dtype="i8")

        return

    def _convert_images(self, grp, images: List[Atoms]):
        """Convert data...

        NOTE:
            support variable number of atoms...

        """
        # - save structures
        nimages = len(images)
        natoms_list = np.array([len(a) for a in images], dtype=np.int32)
        boxes = np.array(
            [a.get_cell(complete=True) for a in images], dtype=np.float64
        ).reshape(-1, 9)
        pbcs = np.array(
            [a.get_pbc() for a in images], dtype=np.int8
        )
        # atomic_numbers = np.array(
        #    [a.get_atomic_numbers() for a in images], dtype=np.int8
        # )
        # positions = np.array([a.get_positions() for a in images], dtype=np.float64)
        atomic_numbers = np.zeros((nimages, max(natoms_list)), dtype=np.int32)
        positions = np.zeros((nimages, max(natoms_list), 3), dtype=np.float64)
        for i, a in enumerate(images):
            atomic_numbers[i, :natoms_list[i]] = a.get_atomic_numbers()
            positions[i, :natoms_list[i], :] = a.get_positions()

        # -- sys props
        natoms_dset = grp.create_dataset(
            "natoms", data=natoms_list, dtype="i8")
        box_dset = grp.create_dataset("box", data=boxes, dtype="f8")
        pbc_dset = grp.create_dataset("pbc", data=pbcs, dtype="i8")
        atype_dset = grp.create_dataset(
            "atype", data=atomic_numbers, dtype="i8")
        pos_dset = grp.create_dataset("positions", data=positions, dtype="f8")

        # - add some info
        #   confid, step
        #   max_devi_e, min_devi_e, avg_devi_e
        #   max_devi_v, min_devi_v, avg_devi_v
        #   max_devi_f, min_devi_f, avg_devi_f
        for name, dtype in zip(RETAINED_INFO_NAMES, RETAIEND_INFO_DTYPES):
            data = [a.info.get(name, np.nan) for a in images]
            if not np.all(np.isnan(data)):
                _ = grp.create_dataset(name, data=data, dtype=dtype)

        # -- calc props
        # TODO: without calc properties?
        # NOTE: energy and forces have apply_constraint True...
        energies = np.array([a.get_potential_energy()
                            for a in images], dtype=np.float64)
        free_energies = []
        for i, a in enumerate(images):
            try:
                free_energy = a.get_potential_energy(force_consistent=True)
            except: # Assume no free_energy property is available.
                free_energy = energies[i]
            free_energies.append(free_energy)

        # forces = np.array([a.get_forces() for a in images], dtype=np.float64)
        forces = np.zeros((nimages, max(natoms_list), 3), dtype=np.float64)
        for i, a in enumerate(images):
            forces[i, :natoms_list[i], :] = a.get_forces()

        ene_dset = grp.create_dataset("energy", data=energies, dtype="f8")
        fen_dset = grp.create_dataset("free_energy", data=free_energies, dtype="f8")
        frc_dset = grp.create_dataset("forces", data=forces, dtype="f8")

        return

    # @classmethod
    # def squeeze(cls, axis=0):
    #    """Squeeze TODO: treat markers and map properly."""

    #    return cls()

    # def take(self, indices, axis=None):
    #    """"""

    #    return

    def __getitem__(self, key):
        """"""
        if isinstance(key, numbers.Integral) or isinstance(key, slice):
            key = [key] + [slice(None) for _ in range(len(self._shape)-1)]
        elif not isinstance(key, tuple):
            raise IndexError("Index must be an integer, a slice or a tuple.")
        assert len(key) <= len(self._shape), "Out of dimension."
        # BUG: <=?
        # print(f"key: {key}")

        # - get indices for each dimension
        indices, tshape = [], []
        for dim, i in enumerate(key):
            size = self._shape[dim]
            if isinstance(i, numbers.Integral):
                if i < -size or i >= size:
                    raise IndexError(
                        f"index {i} is out of bounds for axis {dim} with size {size}.")
                    # IndexError: index 1 is out of bounds for axis 0 with size 1
                if i < 0:
                    i += size
                curr_indices = [i]
            elif isinstance(i, slice):
                curr_indices = range(size)[i]
                for c_i in curr_indices:
                    if c_i <= -size or c_i >= size:
                        raise IndexError(f"Index {c_i} out of range {size}.")
                tshape.append(len(curr_indices))
            else:
                raise IndexError(
                    f"Index must be an integer or a slice for dimension {dim}.")
            indices.append(curr_indices)
        # print(f"tshape: {tshape}")

        # - convert indices
        products = list(itertools.product(*indices))
        global_indices = [_map_idx(x, self._shape) for x in products]
        # print(global_indices)
        # print(self._data)
        # print(f"nframes: {len(self._data)}")
        # print(self._ind_map)

        # - get data
        ret_data = []
        for x in global_indices:
            if x in self._ind_map:
                ret_data.append(self._data[self._ind_map[x]])
            else:
                ret_data.append(None)
        if tshape:
            ret = _reshape_data(ret_data, tshape)
        else:  # tshape is empty, means this is a single atoms
            ret = ret_data[0]

        return ret  # TODO: should this also be an array?

    def __len__(self):
        """"""
        return len(self._data)

    def __repr__(self) -> str:
        """"""

        return f"atoms_array(nimages: {len(self)}, shape: {self.shape})"


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import collections
import itertools

import numpy as np

from ase.io import read, write
from ase.neighborlist import NeighborList, natural_cutoffs

from .gptools import compute_body3_descriptor
from .bench import benchmark


def switch_function(r, r_cut, gamma=2):
    """"""
    pr = r/r_cut

    return 1. - (1.+gamma)*pr**gamma + gamma*pr**(1+gamma)

def switch_function_gradient(r, r_cut, gamma=2):
    """"""
    pr = r/r_cut

    return (-(1.+gamma)*gamma*pr**(gamma-1) + gamma*(1+gamma)*pr**gamma)/r_cut

def switch_function_value_and_grad(r, r_cut, gamma=2):
    """"""
    pr = r/r_cut

    return (
        1. - (1.+gamma)*pr**gamma + gamma*pr**(1+gamma),
        (-(1.+gamma)*gamma*pr**(gamma-1) + gamma*(1+gamma)*pr**gamma)/r_cut
    )


def radial_function(r, r_span, width, r_cut, gamma):
    """Compute radial function.

    Args:
        r:
        r_span: shape (1, num_span)
        f_cut: shape (num_r, 1)

    """
    constant = 1./(r_cut/r_span.shape[1])
    f_cut = switch_function(r, r_cut, gamma)

    return constant*1./r**2*f_cut*np.exp(-(r-r_span)**2/2/width**2)


def radial_function_gradient(r, r_span, width, r_cut):
    """Compute radial function.

    Args:
        r:
        r_span: shape (1, num_span)
        f_cut: shape (num_r, 1)

    """
    constant = 1./(r_cut/r_span.shape[1])

    f_cut, fg_cut = switch_function_value_and_grad(r, r_cut) # r_ij

    v = np.exp(-(r-r_span)**2/2/width**2)

    x = (
        1./r**2*f_cut*(-(r-r_span)/width**2)*v +
        (-2./r**3*f_cut+1./r**2*fg_cut)*v
    )

    return constant*x

def angular_function():
    """"""

    return

def covariance_function(rho1, rho2):
    """"""

    return


def compute_feature():
    """"""
    radial = radial_function(body2_features, r_span, r_delta, r_cut, 2)
    print(radial.shape)
    print(np.sum(radial, axis=0))

    radial_grad = radial_function_gradient(body2_features, r_span, r_delta, r_cut)

    #test_b2 = np.array([[1.28]])
    #radial_grad = radial_function_gradient(test_b2, r_span, r_delta, r_cut)
    print(radial_grad.shape)

    natoms = 4
    radial_gcart = np.zeros((natoms, 3))
    for dimer_index, gcart, gradial in zip(body2_mapping, body2_gradients, radial_grad):
        i, j = dimer_index[1], dimer_index[2]
        radial_gcart[i, :] += np.sum(gradial[:, np.newaxis]*gcart[:, 0:3], axis=0)
        radial_gcart[j, :] += np.sum(gradial[:, np.newaxis]*gcart[:, 3:6], axis=0)
    print(radial_gcart)

    return


def train():
    """"""
    frames = read("./md/Cu4/cand1/traj.xyz", ":1")
    nframes = len(frames)

    print("!!! JAX !!!")
    benchmark(frames)

    print("!!! MANUAL !!!")

    # - 
    r_cut = 6.0
    r_delta = 0.5

    # -
    (
        body2_mapping, body2_features, body2_gradients, 
        body3_mapping, body3_features, body3_gradients
    ) = compute_body3_descriptor(frames, r_cut=r_cut)

    # - b2
    print("- BODY-2 -")
    #print(body2_features)
    print(body2_mapping)

    #r_span = np.linspace(0., r_cut, num=13, endpoint=True)[np.newaxis, :]
    r_span = np.linspace(-1., r_cut+2.0, num=13)[np.newaxis, :]
    num_dim = r_span.shape[1]

    # -
    radial = radial_function(body2_features, r_span, r_delta, r_cut, 2)

    features = np.zeros((nframes, num_dim))
    for dimer_index, v in zip(body2_mapping, radial):
        fi = dimer_index[0]
        features[fi, :] += v
    print(features)

    return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" ase implementation of fort.188
    to sample a specific reactions,
    combine interplotation and overshooting,
    automatically sample IS, FS and intermediate points between them
"""

import numpy as np

import ase
from ase.io import read, write
from ase.constraints import FixBondLength, FixAtoms

from ase.neb import NEB
from ase.optimize import BFGS

from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.langevin import Langevin

from sella import Sella, Constraints

atoms_IS = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-IS.xyz")
atoms_TS = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-TS.xyz")
atoms_FS = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-FS.xyz")

def add_188():
    cons = FixAtoms(indices=[a.index for a in atoms if a.position[2]<7.0])
    cons_bond = FixBondLength(0, 23)

    atoms.set_constraint([cons, cons_bond])

    return

from eann.interface.ase.calculator import Eann
calc = Eann(
    # for now, ase interface only supports double precision
    model = "/mnt/scratch2/users/40247882/catsign/ZnOx/P188/ZnO_200_DOUBLE.pt",
    type_map = {
        "C": 0, "O": 1, "Zn": 2
    }
)

def run_sella():
    """"""
    atoms = atoms_TS
    atoms.calc = calc
    dyn = Sella(
        atoms,
        order = 1,
        internal = False,
        trajectory="opt.traj"
    )
    dyn.run(0.05, 100)

    return

def run_single():
    maxforces = []
    energies = []
    for atoms in [atoms_IS, atoms_TS, atoms_FS]:
        calc.reset()
        atoms.calc = calc
        forces = atoms.get_forces()
        maxforce = np.max(np.fabs(forces))
        maxforces.append(maxforce)
        energies.append(atoms.get_potential_energy())
    energies = np.array(energies)
    print(energies)
    print(energies - energies[0])
    print(maxforces)

    return

def run_NEB():
    initial = atoms_IS
    final = atoms_FS

    nimages = 5
    images = [initial]
    images += [initial.copy() for i in range(nimages-2)]
    images.append(final)

    cons = FixAtoms(indices=[a.index for a in initial if a.position[2]<7.0])

    for atoms in images:
        calc = Eann(
            model = "/mnt/scratch2/users/40247882/catsign/ZnOx/P188/ZnO_200_DOUBLE.pt",
            type_map = {
                "C": 0, "O": 1, "Zn": 2
            }
        )
        atoms.calc = calc
        atoms.set_constraint(cons)

    neb = NEB(
        images, allow_shared_calculator=False,
        k=0.1
        # dynamic_relaxation = False
    )
    #neb = SingleCalculatorNEB(images)
    
    neb.interpolate()
    #print(neb.images)
    
    qn = BFGS(neb, trajectory="neb.traj")
    qn.run(fmax=0.05, steps=100)

    return

def run_opt():
    maxforces = []
    energies = []
    for i, atoms in enumerate([atoms_IS, atoms_TS, atoms_FS]):
        calc.reset()
        atoms.calc = calc
        qn = BFGS(atoms, trajectory=f"opt-{i}.traj")
        qn.run(fmax=0.05, steps=100)
        # info
        forces = atoms.get_forces()
        maxforce = np.max(np.fabs(forces))
        maxforces.append(maxforce)
        energies.append(atoms.get_potential_energy())
    energies = np.array(energies)
    print(energies)
    print(energies - energies[0])
    print(maxforces)

    return

# MD
def run_MD():
    timestep = 2.0
    temperature = 600
    MaxwellBoltzmannDistribution(atoms, temperature_K=temperature)
    dyn = Langevin(
        atoms, 
        timestep = timestep*ase.units.fs, 
        temperature_K = temperature, 
        friction = 0.002, # TODO: what is the unit?
        fixcm = True
    )

    write("traj.xyz", atoms)
    for i in range(10):
        dyn.step()
        write("traj.xyz", atoms, append=True)

if __name__ == "__main__":
    #run_single()
    run_opt()
    #run_NEB()
    #run_sella()
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import Callable, NoReturn, Union

from .. import config

class Variable:

    """Intrinsic, changeable parameter of a graph.
    """

    identifier: str = "vx"

    #: Working directory for the operation.
    _directory: Union[str,pathlib.Path] = pathlib.Path.cwd()

    #: Working status that should be always finished.
    status = "finished"

    #: Standard print function.
    _print: Callable = config._print

    #: Standard debug function.
    _debug: Callable = config._debug

    def __init__(self, initial_value=None, directory="./"):
        """"""
        self.value = initial_value
        self.consumers = []

        self.directory = directory

        #_default_graph.variables.append(self)

        return

    @property
    def directory(self):
        """"""

        return self._directory
    
    @directory.setter
    def directory(self, directory_) -> NoReturn:
        """"""
        self._directory = pathlib.Path(directory_)

        return
    
    def reset(self):
        """Reset node's output and status."""
        if hasattr(self, "output"):
            delattr(self, "output")

        return
    
    def reset_random_seed(self, mode="init"):
        """"""
        if hasattr(self, "_reset_random_seed"):
            self._reset_random_seed(mode=mode)

        return


class DummyVariable(Variable):

    ...


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import logging
import pathlib
import time
import pickle

from typing import NoReturn, List, Mapping

from itertools import combinations, product, chain, groupby

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.formula import Formula

from .. import registers
from .. import StructureBuilder
from .. import ComputerVariable, DriverBasedWorker
from ..expedition import AbstractExpedition

from gdpx.worker.grid import GridDriverBasedWorker
from gdpx.potential.interface import create_mixer
from gdpx.builder.group import create_a_group, create_a_molecule_group
from gdpx.graph.molecule import find_product, find_molecules


def convert_index_to_formula(atoms, group_indices: List[List[int]]):
    """"""
    formulae = []
    for g in group_indices:
        symbols = [atoms[i].symbol for i in g]
        formulae.append(
            Formula.from_list(symbols).format("hill")
        )
    #formulae = sorted(formulae)

    return formulae


def find_target_fragments(atoms, target_commands: List[str]) -> Mapping[str,List[List[int]]]:
    """Find target fragments in the structure to react.

    This is a wrapper for group commands as there are several ways to defind
    a (atomic) group but we need a group of molecules here. Optional ways are 
    using `molecule` all, using `tag`...

    """
    fragments = {} # Mapping[str,List[List[int]]]
    target_molecules = None

    ngroups = len(target_commands)
    if ngroups == 1:
        atomic_indices = create_a_group(atoms, target_commands[0])
        fragments = find_molecules(atoms, atomic_indices)
    else:
        assert ngroups >= 2, "Need at least 2 groups..."
        for group_command in target_commands:
            fragments[group_command] = create_a_molecule_group(atoms, group_command)

    return fragments


class AFIRSearch(AbstractExpedition):


    def __init__(
            self, builder, target, gamma: List[float]=[0.5, 2.5, 1.0], 
            seed=None, directory="./", *args, **kwargs
        ) -> None:
        """Define some basic parameters for the afir search.

        Args:
            builder: StructureBuilder.
        
        """
        self.directory = directory

        self.builder = builder
        self.target = target
        #assert len(self.target) == 2, "Target only supports two elements."
        self.gamma = gamma

        self.bias_params = dict(
            name = "bias", 
            params = dict(
                backend="ase", type = "afir",
                gamma = None, groups = None
            )
        )

        return

    def register_worker(self, worker: dict, *args, **kwargs):
        """"""
        if isinstance(worker, dict):
            worker_params = copy.deepcopy(worker)
            worker = registers.create(
                "variable", "computer", convert_name=True, **worker_params
            ).value[0]
        elif isinstance(worker, list): # assume it is from a computervariable
            worker = worker[0]
        elif isinstance(worker, ComputerVariable):
            worker = worker.value[0]
        elif isinstance(worker, DriverBasedWorker):
            worker = worker
        else:
            raise RuntimeError(f"Unknown worker type {worker}")
        
        self.worker = worker

        return

    def _prepare_fragments(self, atoms, *args, **kwargs):
        """"""
        data_path = self.directory / "_data"
        if not data_path.exists():
            data_path.mkdir(parents=True, exist_ok=True)

        # - find possible reaction pairs
        frag_fpath = data_path/"fragments.pkl"
        # TODO: assure the pair order is the same when restart
        if not frag_fpath.exists():
            fragments = find_target_fragments(atoms, self.target)
            with open(frag_fpath, "wb") as fopen:
                pickle.dump(fragments, fopen)
        else:
            with open(frag_fpath, "rb") as fopen:
                fragments = pickle.load(fopen)

        # TODO: assert molecules in one group are the same type?
        self._print("Found Target Fragments: ")
        for k, v in fragments.items():
            self._print("  {:<24s}:  {}".format(k, v))

        frag_list = []
        for k, v in fragments.items():
            frag_list.append(v)
        
        ntypes = len(frag_list)
        comb = combinations(range(ntypes), 2)

        possible_pairs = []
        for i, j in comb:
            f1, f2 = frag_list[i], frag_list[j]
            possible_pairs.extend(list(product(f1, f2)))
        
        with open(self.directory/"_data"/"pairs.pkl", "wb") as fopen:
            pickle.dump(possible_pairs, fopen)

        return possible_pairs
    
    def _create_afir_potters(self, pair: List[List[int]], gamma_list: List[float], *args, **kwargs):
        """"""
        if hasattr(self.worker.potter, "remove_loaded_models"):
            self.worker.potter.remove_loaded_models()

        bias_list = []
        for g in gamma_list:
            curr_bias = copy.deepcopy(self.bias_params)
            curr_bias["params"]["groups"] = pair 
            curr_bias["params"]["gamma"] = g
            bias_list.append(curr_bias)
        potters = [create_mixer(self.worker.potter, b) for b in bias_list]

        return potters
    
    def run(self, *args, **kwargs) -> None:
        """"""
        # - some imported packages change `logging.basicConfig` 
        #   and accidently add a StreamHandler to logging.root
        #   so remove it...
        for h in logging.root.handlers:
            if isinstance(h, logging.StreamHandler) and not isinstance(h, logging.FileHandler):
                logging.root.removeHandler(h)

        self._print("---------------------------")
        self._print("| AFIRSearch starts...... |")
        self._print("---------------------------")
        # TODO: check restart

        # - assume input structures are minimised
        frames = self.builder.run()
        nframes = len(frames)
        assert nframes == 1, "Only one structure for now."
        atoms = frames[0]

        # - find fragments
        possible_pairs = self._prepare_fragments(atoms)
        
        # - prepare afir bias
        gamma_list = np.linspace(*self.gamma, endpoint=True)

        # - run each pair
        grid_workers = []
        for i, pair in enumerate(possible_pairs):
            # -- start info
            self._print(f"===== Pair {i} =====")
            reactants = convert_index_to_formula(atoms, pair)
            self._print("Reactants:")
            self._print(reactants)
            self._print(pair)

            # -- create GridWorker
            potters = self._create_afir_potters(pair, gamma_list)
            curr_worker = GridDriverBasedWorker(
                potters, driver=self.worker.driver
            )
            curr_worker.directory = self.directory / f"pair{i}"
            grid_workers.append(curr_worker)
            curr_worker.run([atoms])
        
        # - extract each pair
        worker_status = []
        for i, curr_worker in enumerate(grid_workers):
            curr_worker.inspect(resubmit=True)
            if curr_worker.get_number_of_running_jobs() == 0:
                worker_status.append(True)
            else:
                worker_status.append(False)
        
        if all(worker_status):
            self._print("---------------------------")
            self._print("| AFIRSearch is finished. |")
            self._print("---------------------------")
            nimages = len(gamma_list)
            for i, (pair, curr_worker) in enumerate(zip(possible_pairs, grid_workers)):
                self._print(f"===== Pair {i} =====")
                reactants = convert_index_to_formula(atoms, pair)
                self._print("Reactants:")
                self._print(reactants)
                self._print(pair)
                curr_trajs = curr_worker.retrieve(include_retrieved=True)
                pseudo_pathway = [t[-1] for t in curr_trajs]
                end_atoms = pseudo_pathway[-1]
                prod_indices = find_product(end_atoms, pair)
                products = convert_index_to_formula(end_atoms, prod_indices)
                self._print("Products:")
                self._print(products)
                if sorted(reactants) == sorted(products):
                    # TODO: index may change? molecule reconstruct?
                    self._print("nothing happens...")
                else:
                    self._print("reaction happens...")
                    # TODO: postprocess, minimise FS to create a pathway
            # - 
            with open(self.directory/"FINISHED", "w") as fopen:
                fopen.write(
                    f"FINISHED AT {time.asctime( time.localtime(time.time()) )}."
                )
        else:
            self._print("Some calculations are unfinished.")

        return
    
    def read_convergence(self, *args, **kwargs):
        """"""
        converged = False
        if (self.directory/"FINISHED").exists():
            converged = True

        return converged
    
    def get_workers(self, *args, **kwargs):
        """"""
        workers = []
        if hasattr(self.worker.potter, "remove_loaded_models"):
            self.worker.potter.remove_loaded_models()

        pair_data_path = self.directory/"_data"/"pairs.pkl"
        if pair_data_path.exists():
            with open(pair_data_path, "rb") as fopen:
                pairs = pickle.load(fopen)
            wdirs = sorted(
                list(self.directory.glob("pair*")), key=lambda x: int(x.name[4:])
            )
            assert len(pairs) == len(wdirs), f"Inconsistent number of pairs {len(pairs)} and wdirs {len(wdirs)}."

            gamma_list = np.linspace(*self.gamma, endpoint=True)
            for i, pair in enumerate(pairs):
                potters = self._create_afir_potters(pair, gamma_list)
                curr_worker = GridDriverBasedWorker(
                    potters, driver=self.worker.driver
                )
                curr_worker.directory = self.directory / f"pair{i}"
                workers.append(curr_worker)
        else:
            raise FileNotFoundError("There is no cache for pairs.")
        
        return workers


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import importlib
import logging
import warnings

from .. import config


class Register:

    def __init__(self, registry_name):
        self._dict = {}
        self._name = registry_name

    def __setitem__(self, key, value):
        if not callable(value):
            raise Exception(f"Value of a Registry must be a callable!\nValue: {value}")
        if key is None:
            key = value.__name__
        if key in self._dict:
            warnings.warn(
                "Key %s already in registry %s." % (key, self._name), UserWarning
            )
        self._dict[key] = value

    def register(self, target):
        """Decorator to register a function or class."""

        def add(key, value):
            self[key] = value
            return value

        if callable(target):
            return add(None, target)

        return lambda x: add(target, x)

    def __getitem__(self, key):
        if key not in self._dict:
            raise Exception(f"No {key} in {self._dict.keys()}")
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def keys(self):
        """key"""
        return self._dict.keys()

    def __repr__(self) -> str:
        """"""
        content = f"{self._name.upper()}:\n"

        keys = sorted(list(self._dict.keys()))
        nkeys = len(keys)
        ncols = 5
        nrows = int(nkeys / ncols)
        for i in range(nrows):
            content += ("  " + "{:<24s}" * ncols + "\n").format(
                *keys[i * ncols : i * ncols + ncols]
            )

        nrest = nkeys - nrows * ncols
        if nrest > 0:
            content += ("  " + "{:<24s}" * nrest + "\n").format(*keys[nrows * ncols :])

        return content


class registers:

    #: Session operations.
    operation: Register = Register("operation")

    #: Session variables.
    variable: Register = Register("variable")

    #: Session placeholder
    placeholder: Register = Register("placeholder")

    #: Schedulers.
    scheduler: Register = Register("scheduler")

    #: Managers (Potentials).
    manager: Register = Register("manager")

    #: Trainers (Potential Trainers).
    trainer: Register = Register("trainer")

    #: Dataloaders (Datasets).
    dataloader: Register = Register("dataloader")

    #: Regions.
    region: Register = Register("region")

    #: Builders.
    builder: Register = Register("builder")

    #: Bias.
    bias: Register = Register("bias")

    #: Colvars.
    colvar: Register = Register("colvar")

    #: Modifiers.
    modifier: Register = Register("modifier")

    #: Reactors.
    reactor: Register = Register("reactor")

    #: Expeditions.
    expedition: Register = Register("expedition")

    #: Selectors.
    selector: Register = Register("selector")

    #: Describers.
    describer: Register = Register("describer")

    #: Comparators.
    comparator: Register = Register("comparator")

    #: Validators.
    validator: Register = Register("validator")

    def __init__(self):
        raise RuntimeError("Registries is not intended to be instantiated")

    @staticmethod
    def get(mod_name: str, cls_name: str, convert_name: bool = True, *args, **kwargs):
        """Acquire the target class from modules."""
        # - convert the cls_name by the internal convention
        if convert_name:
            # cls_name = cls_name.capitalize() + mod_name.capitalize()
            cls_name = (
                "".join([x.capitalize() for x in cls_name.strip().split("_")])
                + mod_name.capitalize()
            )

        # - get the class
        curr_register = getattr(registers, mod_name)
        target_cls = curr_register[cls_name]

        return target_cls

    @staticmethod
    def create(
        mode_name: str, cls_name: str, convert_name: bool = True, *args, **kwargs
    ):
        """"""
        target_cls = registers.get(mode_name, cls_name, convert_name, *args, **kwargs)
        instance = target_cls(*args, **kwargs)

        return instance


ALL_MODULES = [
    # - working components.
    # -- schedulers
    ("gdpx", ["scheduler"]),
    # -- managers (potentials)
    ("gdpx.potential", ["managers"]),
    # -- dataloaders (datasets)
    ("gdpx.data", ["dataset"]),
    # -- bias
    ("gdpx", ["bias"]),
    # -- builders
    ("gdpx", ["builder"]),
    # -- genetic-algorithm-related
    ("gdpx.builder", ["crossover", "mutation"]),
    # -- colvar
    ("gdpx", ["colvar"]),
    # -- selectors
    ("gdpx", ["selector"]),
    # -- describer
    ("gdpx", ["describer"]),
    # -- comparators
    ("gdpx", ["comparator"]),
    # -- expeditions
    ("gdpx", ["expedition"]),
    # -- reactors
    # -- validators
    ("gdpx", ["validator"]),
    # - session operations + variables.
    ("gdpx.builder", ["interface"]),
    ("gdpx.computation", ["interface"]),
    ("gdpx", ["data"]),
    ("gdpx.data", ["interface"]),
    ("gdpx.describer", ["interface"]),
    ("gdpx.potential", ["interface"]),
    ("gdpx.reactor", ["interface"]),
    ("gdpx.comparator", ["interface"]),
    ("gdpx.trainer", ["interface"]),
    ("gdpx.selector", ["interface"]),
    ("gdpx.scheduler", ["interface"]),
    ("gdpx.validator", ["interface"]),
    ("gdpx.worker", ["interface"]),
]


def _handle_errors(errors):
    """Log out and possibly reraise errors during import."""
    names = []  # unimported module names
    if errors:
        for name, err in errors:
            warnings.warn("Module {} import failed: {}".format(name, err), UserWarning)
            names.append(name)
    else:
        ...

    return names


def import_all_modules_for_register(custom_module_paths=None) -> str:
    """Import all modules for register."""
    modules = []
    for base_dir, submodules in ALL_MODULES:
        for name in submodules:
            full_name = base_dir + "." + name
            modules.append(full_name)
    if isinstance(custom_module_paths, list):
        modules += custom_module_paths
    # print("ALL MODULES: ", modules)
    errors = []
    for module in modules:
        try:
            importlib.import_module(module)
        except ImportError as error:
            errors.append((module, error))
    names = _handle_errors(errors)

    # - some imported packages change `logging.basicConfig`
    #   and accidently add a StreamHandler to logging.root
    #   so remove it...
    for h in logging.root.handlers:
        if isinstance(h, logging.StreamHandler) and not isinstance(
            h, logging.FileHandler
        ):
            logging.root.removeHandler(h)

    keys = sorted(names)
    nkeys = len(keys)
    ncols = 3
    nrows = int(nkeys / ncols)

    lines = ["FAILED TO IMPORT OPTIONAL MODULES: "]
    for i in range(nrows):
        lines.append(
            ("  " + "{:<48s}" * ncols + "").format(*keys[i * ncols : i * ncols + ncols])
        )

    nrest = nkeys - nrows * ncols
    if nrest > 0:
        lines.append(("  " + "{:<48s}" * nrest + "").format(*keys[nrows * ncols :]))

    for line in lines:
        config._print(line)

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib

import numpy as np

from ase.calculators.calculator import all_changes
from ase.calculators.mixing import MixedCalculator, LinearCombinationCalculator

from .. import config as GDPCONFIG

class AddonCalculator(MixedCalculator):

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        """"""
        super().calculate(atoms, properties, system_changes)

        if "forces" in properties:
            forces1 = self.calcs[0].get_property("forces", atoms)
            forces2 = self.calcs[1].get_property("forces", atoms)
            self.results["force_contributions"] = np.hstack([forces1,forces2])

        return


class EnhancedCalculator(LinearCombinationCalculator):

    def __init__(self, calcs, save_host=True, weights=None, atoms=None):
        """Init the enhanced calculator.

        Args:
            calcs: Calculators.
            save_host: Whether save host energy and forces.
        """
        if weights is None:
            weights = np.ones(len(calcs))
        super().__init__(calcs, weights, atoms)

        self.save_host = save_host

        return

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        """"""
        # - for sub calculators...
        for i, subcalc in enumerate(self.calcs):
            subcalc.directory = str(
                (pathlib.Path(self.directory)/(str(i).zfill(2)+"."+subcalc.__class__.__name__)).resolve()
            )
        # NOTE: nequip requires that atoms has NequipCalculator or None
        #       thus, we set atoms.calc to None and restore it later
        prev_calc = atoms.calc
        atoms.calc = None

        super().calculate(atoms, properties, system_changes)
        atoms.calc = prev_calc

        if self.save_host:
            self.results["host_energy"] = self.calcs[0].get_property("energy", atoms)
            self.results["host_forces"] = self.calcs[0].get_property("forces", atoms)
        
        # - save deviation if the host calculator is a committee
        if isinstance(self.calcs[0], CommitteeCalculator):
            natoms = len(atoms)
            for k, v in self.calcs[0].results.items():
                if k in GDPCONFIG.VALID_DEVI_FRAME_KEYS:
                    self.results[k] = v
            for k, v in self.calcs[0].results.items():
                if k in GDPCONFIG.VALID_DEVI_ATOMIC_KEYS:
                    self.results[k] = np.reshape(v, (natoms, -1))

        return
    

class CommitteeCalculator(LinearCombinationCalculator):

    def __init__(self, calcs, use_avg=False, save_atomic=True, ddof=0, atoms=None):
        """Init the committee calculator.

        Args:
            calcs: ASE calculators.
            use_avg: Whether use average results instead of those by the first calc.
            save_atomic: Whether save atomic deviation.
            ddof: Dela Degrees of Freedom that affects the deviations of results.

        """
        weights = np.ones(len(calcs))
        if use_avg:
            weights = weights / np.sum(weights)
        else:
            weights[1:] = 0.
        self.ddof = ddof
        self.save_atomic = save_atomic

        super().__init__(calcs, weights, atoms)

        return

    def calculate(self, atoms=None, properties=["energy"], system_changes=all_changes):
        """"""
        # NOTE: nequip requires that atoms has NequipCalculator or None
        #       thus, we set atoms.calc to None and restore it later
        prev_calc = atoms.calc
        atoms.calc = None

        # TODO: set wdir for each subcalc
        super().calculate(atoms, properties, system_changes)
        atoms.calc = prev_calc

        self._compute_deviation(atoms, properties)

        return
    
    def _compute_deviation(self, atoms, properties):
        """Compute the RMSE deviation of calculator properties."""
        if "energy" in properties:
            tot_energies = np.array([c.results["energy"] for c in self.calcs])
            self.results["devi_te"] = np.sqrt(np.var(tot_energies, ddof=self.ddof))

        if "forces" in properties:
            cmt_forces = np.array([c.results["forces"].flatten() for c in self.calcs])
            frc_devi = np.sqrt(np.var(np.array(cmt_forces), axis=0))
            self.results["max_devi_f"] = np.max(frc_devi)
            self.results["min_devi_f"] = np.min(frc_devi)
            self.results["avg_devi_f"] = np.mean(frc_devi)
            if self.save_atomic:
                self.results["devi_f"] = np.reshape(frc_devi, (-1,3))
        
        # TODO: atomic energies?

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import abc

import numpy as np

from ..core.node import AbstractNode


class AbstractComparator(AbstractNode):

    ...

    def compare_composition(self, a1, a2):
        """"""
        # TODO: compare PBC?
        is_similar = False
        na1, na2 = len(a1), len(a2)
        if na1 == na2:
            c1, c2 = a1.get_cell(complete=True), a2.get_cell(complete=True)
            if np.allclose(c1, c2):
                s1, s2 = a1.get_chemical_formula(), a2.get_chemical_formula()
                if s1 == s2:
                    is_similar = True
            else:
                ...
        else:
            ...

        return is_similar

    def __call__(self, a1, a2) -> bool:
        """"""

        return self.looks_like(a1, a2)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .bond import BondGaussianCalculator
from .com import CenterOfMassGaussianCalculator
from .rmsd import RMSDGaussian


__all__ = [
    "BondGaussianCalculator",
    "CenterOfMassGaussianCalculator",
    "RMSDGaussian"
]


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import abc
import dataclasses
import pathlib
from typing import Union

import numpy as np

from ase import Atoms

from ..core.node import AbstractNode


"""Find possible reaction pathways in given structures.
"""


class AbstractReactor(AbstractNode):

    """Base class of an arbitrary reactor.

    A valid reactor may contain the following components: 
        - A builder that offers input structures
        - A worker that manages basic dynamics task (minimisation and MD)
            - driver with two calculator for PES and BIAS
            - scheduler
        - A miner that finds saddle points and MEPs
    and the results would be trajectories and pathways for further analysis.

    """

    def __init__(self, calc, directory: Union[str, pathlib.Path] = "./", random_seed: int = None, *args, **kwargs):
        """"""
        super().__init__(directory, random_seed, *args, **kwargs)
        self.calc = calc

        return
    
    def reset(self):
        """"""
        self.calc.reset()

        return

    @abc.abstractmethod
    def run(self, structures, read_cache: bool=True, *args, **kwargs):
        """"""
        if not self.directory.exists():
            self.directory.mkdir(parents=True)
        
        return 
    
    @abc.abstractmethod
    def read_convergence(self, *args, **kwargs) -> bool:
        """"""
        converged = False

        return converged


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import List

import numpy as np

from ase import Atoms
from ase import data, units
from ase.neighborlist import NeighborList, natural_cutoffs

from .operator import AbstractOperator


class MoveOperator(AbstractOperator):

    name: str = "move"

    def __init__(
        self,
        particles: List[str] = None,
        region: dict = {},
        temperature: float = 300.0,
        pressure: float = 1.0,
        covalent_ratio=[0.8, 2.0],
        max_disp: float = 2.0,
        use_rotation: bool = True,
        *args,
        **kwargs,
    ):
        """"""
        super().__init__(
            region=region,
            temperature=temperature,
            pressure=pressure,
            covalent_ratio=covalent_ratio,
            use_rotation=use_rotation,
            *args,
            **kwargs,
        )

        self.particles = particles

        self.max_disp = max_disp

        return

    def run(self, atoms: Atoms, rng=np.random) -> Atoms:
        """"""
        super().run(atoms)
        self._extra_info = "-"

        # BUG: If there is no species in the system...
        species_indices = self._select_species(atoms, self.particles, rng=rng)

        # - basic
        curr_atoms = atoms.copy()  # TODO: use clean atoms?
        cell = curr_atoms.get_cell(complete=True)

        # - neighbour list
        nl = NeighborList(
            self.covalent_max * np.array(natural_cutoffs(curr_atoms)),
            skin=0.0,
            self_interaction=False,
            bothways=True,
        )

        # - find tag atoms
        # record original position of species_indices
        species = curr_atoms[species_indices]
        self._extra_info = f"Move_{species.get_chemical_formula()}_{species_indices}"

        # org_pos = new_atoms[species_indices].position.copy() # original position
        # TODO: deal with pbc, especially for move step
        org_com = np.mean(species.positions, axis=0)
        org_positions = species.positions.copy()

        # - move the atom
        for i in range(self.MAX_RANDOM_ATTEMPTS):
            rsq = 1.1
            while rsq > 1.0:
                rvec = 2 * rng.uniform(size=3) - 1.0
                rsq = np.linalg.norm(rvec)
            ran_pos = org_com + rvec * self.max_disp
            # -- make a copy and rotate
            species_ = self._rotate_species(species, rng=rng)
            curr_cop = np.average(species_.positions, axis=0)
            # -- translate
            new_vec = ran_pos - curr_cop
            species_.translate(new_vec)
            curr_atoms.positions[species_indices] = species_.positions.copy()
            # use neighbour list
            if not self.check_overlap_neighbour(nl, curr_atoms, cell, species_indices):
                self._print(f"succeed to random after {i+1} attempts...")
                self._print(f"original position: {org_com}")
                self._print(f"random position: {ran_pos}")
                self._print(
                    f"actual position: {np.average(curr_atoms.positions[species_indices], axis=0)}"
                )
                break
            curr_atoms.positions[species_indices] = org_positions
        else:
            curr_atoms = None

        return curr_atoms

    def metropolis(self, prev_ene: float, curr_ene: float, rng=np.random) -> bool:
        """"""
        # - acceptance ratio
        kBT_eV = units.kB * self.temperature
        beta = 1.0 / kBT_eV  # 1/(kb*T), eV

        coef = 1.0
        ene_diff = curr_ene - prev_ene
        acc_ratio = np.min([1.0, coef * np.exp(-beta * (ene_diff))])

        # content = "\nVolume %.4f Nexatoms %.4f CubicWave %.4f Coefficient %.4f\n" %(
        #    self.acc_volume, len(self.tag_list[expart]), cubic_wavelength, coef
        # )
        content = "\nVolume %.4f Beta %.4f Coefficient %.4f\n" % (
            self.region.get_volume(),
            beta,
            coef,
        )
        content += "Energy Difference %.4f [eV]\n" % ene_diff
        content += "Accept Ratio %.4f\n" % acc_ratio
        for x in content.split("\n"):
            self._print(x)

        rn_move = rng.uniform()
        self._print(f"{self.__class__.__name__} Probability %.4f" % rn_move)

        return rn_move < acc_ratio

    def as_dict(self) -> dict:
        """"""
        params = super().as_dict()
        params["particles"] = self.particles
        params["max_disp"] = self.max_disp

        return params

    def __repr__(self) -> str:
        """"""
        content = f"@Modifier {self.__class__.__name__}\n"
        content += (
            f"temperature {self.temperature} [K] pressure {self.pressure} [bar]\n"
        )
        content += "covalent ratio: \n"
        content += f"  min: {self.covalent_min} max: {self.covalent_max}\n"
        content += f"max disp: {self.max_disp}\n"
        content += f"particles: \n"
        content += f"  {self.particles}\n"

        return content


if __name__ == "__main__":
    ...
